from __future__ import annotations

from collections import namedtuple
from copy import copy
from random import randint, uniform

import numpy as np
import pytest

from scitbx import matrix

from dials.algorithms.profile_model.ellipsoid.parameterisation import (
    Angular2MosaicityParameterisation,
    Angular4MosaicityParameterisation,
    ModelState,
    ReflectionModelState,
    Simple1MosaicityParameterisation,
    Simple6MosaicityParameterisation,
    WavelengthSpreadParameterisation,
)
from dials.algorithms.refinement.parameterisation.crystal_parameters import (
    CrystalOrientationParameterisation,
    CrystalUnitCellParameterisation,
)
from dials.array_family import flex


def test_Simple1MosaicityParameterisation():

    p = Simple1MosaicityParameterisation(params=np.array([1e-3]))

    assert p.is_angular() is False
    assert p.num_parameters() == 1
    assert p.parameters == (1e-3,)
    p.parameters = np.array([2e-3])
    assert p.parameters == (2e-3,)
    psq = p.parameters[0] ** 2
    assert list(p.sigma().flatten()) == pytest.approx([psq, 0, 0, 0, psq, 0, 0, 0, psq])
    d = p.first_derivatives()
    assert d.shape[0] == 1
    d1 = 2 * p.parameters[0]
    assert list(d[0, :, :].flatten()) == pytest.approx([d1, 0, 0, 0, d1, 0, 0, 0, d1])


def test_Simple6MosaicityParameterisation():

    params = np.array([1e-3, 2e-3, 3e-3, 4e-3, 5e-3, 6e-3])

    p = Simple6MosaicityParameterisation(params=params)

    assert p.is_angular() is False
    assert p.num_parameters() == 6
    assert list(p.parameters) == pytest.approx(list(params))

    params = np.array([2e-3, 3e-3, 4e-3, 5e-3, 6e-3, 7e-3])
    p.parameters = params
    assert list(p.parameters) == pytest.approx(list(params))

    b1, b2, b3, b4, b5, b6 = params
    assert p.sigma()[0, 0] == pytest.approx(b1**2)
    assert p.sigma()[0, 1] == pytest.approx(b1 * b2)
    assert p.sigma()[0, 2] == pytest.approx(b1 * b4)
    assert p.sigma()[1, 0] == pytest.approx(b1 * b2)
    assert p.sigma()[1, 1] == pytest.approx(b2**2 + b3 * b3)
    assert p.sigma()[1, 2] == pytest.approx(b2 * b4 + b3 * b5)
    assert p.sigma()[2, 0] == pytest.approx(b1 * b4)
    assert p.sigma()[2, 1] == pytest.approx(b2 * b4 + b3 * b5)
    assert p.sigma()[2, 2] == pytest.approx(b4**2 + b5**2 + b6**2)

    dSdb = [
        (2 * b1, b2, b4, b2, 0, 0, b4, 0, 0),
        (0, b1, 0, b1, 2 * b2, b4, 0, b4, 0),
        (0, 0, 0, 0, 2 * b3, b5, 0, b5, 0),
        (0, 0, b1, 0, 0, b2, b1, b2, 2 * b4),
        (0, 0, 0, 0, 0, b3, 0, b3, 2 * b5),
        (0, 0, 0, 0, 0, 0, 0, 0, 2 * b6),
    ]

    d = p.first_derivatives()
    assert d.shape[0] == 6
    for i in range(d.shape[0]):
        a = dSdb[i]
        b = d[i, :, :]
        for j in range(9):
            assert b.flatten()[j] == pytest.approx(a[j], abs=1e-12)


def test_WavelengthSpreadParameterisation():

    params = np.array([1e-3])

    p = WavelengthSpreadParameterisation(params=params)
    assert p.num_parameters() == 1
    assert p.parameters[0] == pytest.approx(params[0])
    params = np.array([2e-3])
    p.parameters = params
    assert p.parameters[0] == pytest.approx(params[0])
    assert p.sigma() == pytest.approx(params[0] ** 2)
    assert p.first_derivatives()[0] == pytest.approx(2 * params[0])


def test_Angular2MosaicityParameterisation():
    params = np.array([1e-3, 2e-3])

    p = Angular2MosaicityParameterisation(params=params)

    assert p.is_angular() is True
    assert p.num_parameters() == 2
    assert p.parameters[0] == pytest.approx(params[0])
    assert p.parameters[1] == pytest.approx(params[1])

    params = np.array([2e-3, 3e-3])
    p.parameters = params
    assert p.parameters[0] == pytest.approx(params[0])
    assert p.parameters[1] == pytest.approx(params[1])

    b1, b2 = params
    assert list(p.sigma().flatten()) == pytest.approx(
        [b1**2, 0, 0, 0, b1**2, 0, 0, 0, b2**2]
    )

    dSdb = [(2 * b1, 0, 0, 0, 2 * b1, 0, 0, 0, 0), (0, 0, 0, 0, 0, 0, 0, 0, 2 * b2)]

    d = p.first_derivatives()
    assert d.shape[0] == 2
    for i in range(d.shape[0]):
        a = dSdb[i]
        b = d[i, :, :]
        for j in range(9):
            assert b.flatten()[j] == pytest.approx(a[j])


def test_Angular4MosaicityParameterisation():
    params = np.array([1e-3, 2e-3, 3e-3, 4e-3])

    p = Angular4MosaicityParameterisation(params=params)

    assert p.is_angular() is True
    assert p.num_parameters() == 4
    assert list(p.parameters) == pytest.approx(list(params))

    params = np.array([2e-3, 3e-3, 4e-3, 5e-3])
    p.parameters = params
    assert list(p.parameters) == pytest.approx(list(params))

    b1, b2, b3, b4 = params
    assert p.sigma()[0, 0] == pytest.approx(b1**2)
    assert p.sigma()[0, 1] == pytest.approx(b1 * b2)
    assert p.sigma()[0, 2] == pytest.approx(0)
    assert p.sigma()[1, 0] == pytest.approx(b1 * b2)
    assert p.sigma()[1, 1] == pytest.approx(b2**2 + b3 * b3)
    assert p.sigma()[1, 2] == pytest.approx(0)
    assert p.sigma()[2, 0] == pytest.approx(0)
    assert p.sigma()[2, 1] == pytest.approx(0)
    assert p.sigma()[2, 2] == pytest.approx(b4**2)

    dSdb = [
        (2 * b1, b2, 0, b2, 0, 0, 0, 0, 0),
        (0, b1, 0, b1, 2 * b2, 0, 0, 0, 0),
        (0, 0, 0, 0, 2 * b3, 0, 0, 0, 0),
        (0, 0, 0, 0, 0, 0, 0, 0, 2 * b4),
    ]

    d = p.first_derivatives()
    assert d.shape[0] == 4
    for i in range(d.shape[0]):
        a = dSdb[i]
        b = d[i, ::, :]
        for j in range(9):
            assert b.flatten()[j] == pytest.approx(a[j])


def check_model_state_with_fixed(
    experiment,
    mosaicity_parameterisation,
    wavelength_parameterisation,
    fix_mosaic_spread=False,
    fix_wavelength_spread=False,
    fix_unit_cell=False,
    fix_orientation=False,
):

    state = ModelState(
        experiment,
        mosaicity_parameterisation,
        wavelength_parameterisation,
        fix_mosaic_spread=fix_mosaic_spread,
        fix_wavelength_spread=fix_wavelength_spread,
        fix_unit_cell=fix_unit_cell,
        fix_orientation=fix_orientation,
    )

    assert state.is_orientation_fixed == fix_orientation
    assert state.is_unit_cell_fixed == fix_unit_cell
    assert state.is_mosaic_spread_fixed == fix_mosaic_spread
    assert state.is_wavelength_spread_fixed == fix_wavelength_spread

    U = state.U_matrix
    B = state.B_matrix
    A = state.A_matrix
    M = state.mosaicity_covariance_matrix
    L = state.wavelength_spread

    assert U.shape == (3, 3)
    assert B.shape == (3, 3)
    assert A.shape == (3, 3)
    assert M.shape == (3, 3)
    if wavelength_parameterisation is not None:
        assert len(L) == 1
    else:
        assert len(L) == 0

    assert len(state.U_params) == 3
    assert len(state.B_params) == 2
    assert len(state.M_params) == mosaicity_parameterisation.num_parameters()
    if wavelength_parameterisation is not None:
        assert len(state.L_params) == 1
    else:
        assert len(state.L_params) == 0

    dU = state.dU_dp
    dB = state.dB_dp
    dM = state.dM_dp
    dL = state.dL_dp
    assert dU.shape[0] == 3
    assert dB.shape[0] == 2
    assert dM.shape[0] == mosaicity_parameterisation.num_parameters()
    if wavelength_parameterisation is not None:
        assert len(dL) == 1
    else:
        assert len(dL) == 0

    params = state.active_parameters

    expected_len = 0
    if not fix_mosaic_spread:
        expected_len += mosaicity_parameterisation.num_parameters()
    if not fix_wavelength_spread:
        if wavelength_parameterisation is not None:
            expected_len += wavelength_parameterisation.num_parameters()
    if not fix_unit_cell:
        expected_len += 2
    if not fix_orientation:
        expected_len += 3

    assert len(params) == expected_len
    new_params = params
    state.active_parameters = new_params


def test_ModelState(test_experiment):

    experiments = [test_experiment]

    S1 = Simple1MosaicityParameterisation()
    S6 = Simple6MosaicityParameterisation()
    W = WavelengthSpreadParameterisation()

    with pytest.raises(AssertionError):
        check_model_state_with_fixed(experiments[0], S1, None, fix_mosaic_spread=True)
    with pytest.raises(AssertionError):
        check_model_state_with_fixed(experiments[0], S1, None, fix_unit_cell=True)
    with pytest.raises(AssertionError):
        check_model_state_with_fixed(experiments[0], S1, None, fix_orientation=True)
    check_model_state_with_fixed(experiments[0], S1, None, fix_wavelength_spread=True)
    check_model_state_with_fixed(experiments[0], S1, W, fix_mosaic_spread=True)
    check_model_state_with_fixed(experiments[0], S1, W, fix_wavelength_spread=True)
    check_model_state_with_fixed(experiments[0], S1, W, fix_unit_cell=True)
    check_model_state_with_fixed(experiments[0], S1, W, fix_orientation=True)

    with pytest.raises(AssertionError):
        check_model_state_with_fixed(experiments[0], S6, None, fix_mosaic_spread=True)
    with pytest.raises(AssertionError):
        check_model_state_with_fixed(experiments[0], S6, None, fix_unit_cell=True)
    with pytest.raises(AssertionError):
        check_model_state_with_fixed(experiments[0], S6, None, fix_orientation=True)
    check_model_state_with_fixed(experiments[0], S6, None, fix_wavelength_spread=True)
    check_model_state_with_fixed(experiments[0], S6, W, fix_mosaic_spread=True)
    check_model_state_with_fixed(experiments[0], S6, W, fix_wavelength_spread=True)
    check_model_state_with_fixed(experiments[0], S6, W, fix_unit_cell=True)
    check_model_state_with_fixed(experiments[0], S6, W, fix_orientation=True)


def check_reflection_model_state_with_fixed(
    experiment,
    mosaicity_parameterisation,
    wavelength_parameterisation,
    fix_mosaic_spread=False,
    fix_wavelength_spread=False,
    fix_unit_cell=False,
    fix_orientation=False,
):

    state = ModelState(
        experiment,
        mosaicity_parameterisation,
        wavelength_parameterisation,
        fix_mosaic_spread=fix_mosaic_spread,
        fix_wavelength_spread=fix_wavelength_spread,
        fix_unit_cell=fix_unit_cell,
        fix_orientation=fix_orientation,
    )

    model = ReflectionModelState(
        state, matrix.col(experiment.beam.get_s0()), matrix.col((1, 1, 1))
    )

    assert list(model.mosaicity_covariance_matrix.flatten()) == list(
        mosaicity_parameterisation.sigma().flatten()
    )
    assert list(model.get_r().flatten()) == pytest.approx(
        np.matmul(state.A_matrix, np.array([1, 1, 1]).reshape(3, 1))[:, 0].tolist(),
        abs=1e-6,
    )

    if wavelength_parameterisation is not None:
        assert model.wavelength_spread == wavelength_parameterisation.sigma()
    else:
        assert model.wavelength_spread == 0

    dS_dp = model.get_dS_dp()
    dr_dp = model.get_dr_dp()
    dL_dp = model.get_dL_dp()

    assert dS_dp.shape[2] == len(state.parameter_labels)
    assert dr_dp.shape[1] == len(state.parameter_labels)
    assert len(dL_dp) == len(state.parameter_labels)

    if not fix_wavelength_spread:
        assert dr_dp[:, -1].flatten() == pytest.approx([0, 0, 0], abs=1e-6)
        assert dS_dp[:, :, -1].flatten() == pytest.approx(
            (0, 0, 0, 0, 0, 0, 0, 0, 0), abs=1e-6
        )
        dr_dp = dr_dp[:, :-1]
        dS_dp = dS_dp[:, :, :-1]
        dL_dp = dL_dp[:-1]
    if not fix_mosaic_spread:
        num_params = mosaicity_parameterisation.num_parameters()
        for i in range(num_params):
            assert dr_dp[:, -(i + 1)] == pytest.approx([0, 0, 0], abs=1e-6)
            assert dS_dp[:, :, -(i + 1)] == pytest.approx(
                state.dM_dp[-(i + 1), :, :], abs=1e-6
            )
            assert dL_dp[-1] == 0
        dr_dp = dr_dp[:, :-num_params]
        dS_dp = dS_dp[:, :, :-num_params]
        dL_dp = dL_dp[:-num_params]

    if not fix_unit_cell:
        num_params = state.B_params.size
        for i in range(num_params):
            assert dS_dp[:, :, -(i + 1)].flatten() == pytest.approx(
                (0, 0, 0, 0, 0, 0, 0, 0, 0), abs=1e-6
            )
            assert dL_dp[-(i + 1)] == 0
        dr_dp = dr_dp[:, :-num_params]
        dS_dp = dS_dp[:, :, :-num_params]
        dL_dp = dL_dp[:-num_params]

    if not fix_orientation:
        num_params = state.U_params.size
        for i in range(num_params):
            assert dS_dp[:, :, -(i + 1)].flatten() == pytest.approx(
                (0, 0, 0, 0, 0, 0, 0, 0, 0), abs=1e-6
            )
            assert dL_dp[-(i + 1)] == 0
        dr_dp = dr_dp[:, :-num_params]
        dS_dp = dS_dp[:, :, :-num_params]
        dL_dp = dL_dp[:-num_params]


def test_ReflectionModelState(test_experiment):

    experiments = [test_experiment]

    S1 = Simple1MosaicityParameterisation()
    S6 = Simple6MosaicityParameterisation()
    W = WavelengthSpreadParameterisation()

    check_reflection_model_state_with_fixed(
        experiments[0], S1, None, fix_wavelength_spread=True
    )
    check_reflection_model_state_with_fixed(
        experiments[0], S1, W, fix_mosaic_spread=True
    )
    check_reflection_model_state_with_fixed(
        experiments[0], S1, W, fix_wavelength_spread=True
    )
    check_reflection_model_state_with_fixed(experiments[0], S1, W, fix_unit_cell=True)
    check_reflection_model_state_with_fixed(experiments[0], S1, W, fix_orientation=True)

    check_reflection_model_state_with_fixed(
        experiments[0], S6, None, fix_wavelength_spread=True
    )
    check_reflection_model_state_with_fixed(
        experiments[0], S6, W, fix_mosaic_spread=True
    )
    check_reflection_model_state_with_fixed(
        experiments[0], S6, W, fix_wavelength_spread=True
    )
    check_reflection_model_state_with_fixed(experiments[0], S6, W, fix_unit_cell=True)
    check_reflection_model_state_with_fixed(experiments[0], S6, W, fix_orientation=True)


def first_derivative(func, x, h):
    return (-func(x + 2 * h) + 8 * func(x + h) - 8 * func(x - h) + func(x - 2 * h)) / (
        12 * h
    )


def generate_data(experiments, reflections):

    from random import seed

    seed(0)

    index = randint(0, len(reflections))

    h = reflections[index]["miller_index"]

    s0 = matrix.col(experiments[0].beam.get_s0())

    U_param = CrystalOrientationParameterisation(experiments[0].crystal)
    B_param = CrystalUnitCellParameterisation(experiments[0].crystal)

    U = matrix.sqr(experiments[0].crystal.get_U())
    B = matrix.sqr(experiments[0].crystal.get_B())
    r = U * B * matrix.col(h)
    s2 = s0 + r
    mobs = (
        s2 + matrix.col((uniform(0, 1e-3), uniform(0, 1e-3), uniform(0, 1e-3)))
    ).normalize() * s0.length()

    b1, b2, b3, b4, b5, b6 = (
        uniform(1e-3, 3e-3),
        uniform(0.0, 1e-3),
        uniform(1e-3, 3e-3),
        uniform(0.0, 1e-3),
        uniform(0.0, 1e-3),
        uniform(1e-3, 3e-3),
    )

    S_param = (b1, b2, b3, b4, b5, b6)
    L_param = (uniform(1e-3, 2e-3),)
    ctot = randint(100, 1000)

    T = matrix.sqr((uniform(1e-3, 2e-3), 0, uniform(1e-6, 2e-6), uniform(1e-3, 2e-3)))
    Sobs = T * T.transpose()

    params = [S_param, U_param, B_param, L_param]

    return params, s0, h, ctot, mobs, Sobs


@pytest.fixture
def testdata(test_experiment):

    TestData = namedtuple(
        "TestData", ["experiment", "models", "s0", "h", "ctot", "mobs", "Sobs"]
    )

    experiments = [test_experiment]
    reflections = flex.reflection_table.from_predictions_multi(experiments)

    models, s0, h, ctot, mobs, Sobs = generate_data(experiments, reflections)

    return TestData(
        experiment=experiments[0],
        models=models,
        s0=s0,
        h=h,
        ctot=ctot,
        mobs=mobs,
        Sobs=Sobs,
    )


def test_ReflectionModelState_derivatives(testdata):
    def check(
        mosaicity_parameterisation,
        wavelength_parameterisation,
        fix_mosaic_spread=False,
        fix_wavelength_spread=False,
        fix_unit_cell=False,
        fix_orientation=False,
    ):
        experiment = testdata.experiment
        models = testdata.models
        s0 = testdata.s0
        h = testdata.h
        # ctot = testdata.ctot
        # mobs = testdata.mobs
        # Sobs = testdata.Sobs

        U_params = models[1].get_param_vals()
        B_params = models[2].get_param_vals()
        M_params = np.array(models[0][: mosaicity_parameterisation.num_parameters()])
        L_params = np.array(models[3])

        state = ModelState(
            experiment,
            mosaicity_parameterisation,
            wavelength_parameterisation,
            fix_mosaic_spread=fix_mosaic_spread,
            fix_wavelength_spread=fix_wavelength_spread,
            fix_unit_cell=fix_unit_cell,
            fix_orientation=fix_orientation,
        )
        state.U_params = U_params
        state.B_params = B_params
        state.M_params = M_params
        state.L_params = L_params

        model = ReflectionModelState(state, s0, h)

        dr_dp = model.get_dr_dp()
        dS_dp = model.get_dS_dp()
        dL_dp = model.get_dL_dp()

        def compute_sigma(parameters):
            state.active_parameters = parameters
            model = ReflectionModelState(state, s0, h)
            return model.mosaicity_covariance_matrix

        def compute_r(parameters):
            state.active_parameters = parameters
            model = ReflectionModelState(state, s0, h)
            return model.get_r()

        def compute_sigma_lambda(parameters):
            state.active_parameters = parameters
            model = ReflectionModelState(state, s0, h)
            return model.wavelength_spread

        step = 1e-6

        parameters = copy(state.active_parameters)

        dr_num = []
        for i in range(len(parameters)):

            def f(x):
                p = copy(parameters)
                p[i] = x
                return compute_r(p)

            dr_num.append(first_derivative(f, parameters[i], step).reshape(3, 1))
        dr_num = np.concatenate(dr_num, axis=1)

        for n, c in zip(dr_num, dr_dp):
            for nn, cc in zip(n, c):
                print(nn)
                print(cc)
                assert abs(nn - cc) < 1e-7

        ds_num = []
        for i in range(len(parameters)):

            def f(x):
                p = copy(parameters)
                p[i] = x
                return compute_sigma(p)

            fd = first_derivative(f, parameters[i], step)
            print(fd)
            ds_num.append(fd.reshape(3, 3, 1))
        ds_num = np.concatenate(ds_num, axis=2)

        for i in range(len(parameters)):
            for n, c in zip(ds_num[:, :, i], dS_dp[:, :, i]):
                for nn, cc in zip(n.flatten(), c.flatten()):
                    print(nn)
                    print(cc)
                    assert abs(nn - cc) < 1e-5

        dl_num = []
        for i in range(len(parameters)):

            def f(x):
                p = copy(parameters)
                p[i] = x
                return compute_sigma_lambda(p) ** 2

            dl_num.append(first_derivative(f, parameters[i], step))

        for n, c in zip(dl_num, dL_dp):
            assert abs(n - c) < 1e-7

    S1 = Simple1MosaicityParameterisation()
    S6 = Simple6MosaicityParameterisation()
    W = WavelengthSpreadParameterisation()

    check(S1, None, fix_wavelength_spread=True)
    check(S1, W, fix_mosaic_spread=True)
    check(S1, W, fix_wavelength_spread=True)
    check(S1, W, fix_unit_cell=True)
    check(S1, W, fix_orientation=True)

    check(S6, None, fix_wavelength_spread=True)
    check(S6, W, fix_mosaic_spread=True)
    check(S6, W, fix_wavelength_spread=True)
    check(S6, W, fix_unit_cell=True)
    check(S6, W, fix_orientation=True)



from __future__ import annotations

import pickle

import pytest

from dxtbx.model import ExperimentList
from libtbx import phil

from dials.command_line.dials_import import do_import
from dials.command_line.dials_import import phil_scope as import_phil_scope
from dials.command_line.generate_mask import generate_mask, phil_scope


@pytest.fixture(
    params=[
        {
            "directory": "centroid_test_data",
            "filename": "imported_experiments.json",
            "masks": ["pixels.mask"],
        },
        {
            "directory": "l_cysteine_dials_output",
            "filename": "imported.expt",
            "masks": ["pixels_%d.mask" % (i + 1) for i in range(4)],
        },
    ],
    ids=["One sequence", "Four sequences"],
)
def experiments_masks(request, dials_data):
    filename = (
        dials_data(request.param["directory"], pathlib=True) / request.param["filename"]
    )
    return ExperimentList.from_file(filename), request.param["masks"]


def test_generate_mask(experiments_masks, run_in_tmp_path):
    experiments, masks = experiments_masks

    params = phil_scope.fetch().extract()
    generate_mask(experiments, params)

    assert all(run_in_tmp_path.joinpath(mask).is_file() for mask in masks)


def test_generate_mask_with_untrusted_rectangle(experiments_masks, tmpdir):
    experiments, masks = experiments_masks

    params = phil_scope.fetch(
        phil.parse("untrusted.rectangle=100,200,100,200")
    ).extract()
    params.output.experiments = "masked.expt"
    with tmpdir.as_cwd():
        generate_mask(experiments, params)

    assert all(tmpdir.join(mask).check() for mask in masks)
    assert tmpdir.join("masked.expt").check()

    experiments = ExperimentList.from_file(tmpdir.join("masked.expt").strpath)
    imageset = experiments.imagesets()[0]
    assert imageset.external_lookup.mask.filename == tmpdir.join(masks[0]).strpath


def test_generate_mask_with_untrusted_circle(experiments_masks, tmpdir):
    experiments, masks = experiments_masks

    params = phil_scope.fetch(phil.parse("untrusted.circle=100,100,10")).extract()
    with tmpdir.as_cwd():
        generate_mask(experiments, params)

    assert all(tmpdir.join(mask).check() for mask in masks)


def test_generate_mask_with_resolution_range(experiments_masks, tmpdir):
    experiments, masks = experiments_masks

    params = phil_scope.fetch().extract()
    params.resolution_range = [(2, 3)]
    with tmpdir.as_cwd():
        generate_mask(experiments, params)

    assert all(tmpdir.join(mask).check() for mask in masks)


def test_generate_mask_with_d_min_d_max(experiments_masks, tmpdir):
    experiments, masks = experiments_masks

    params = phil_scope.fetch().extract()
    params.d_min = 3
    params.d_max = 2
    with tmpdir.as_cwd():
        generate_mask(experiments, params)

    assert all(tmpdir.join(mask).check() for mask in masks)


def test_generate_mask_with_ice_rings(experiments_masks, tmpdir):
    experiments, masks = experiments_masks

    params = phil_scope.fetch().extract()
    params.ice_rings.filter = True
    params.ice_rings.d_min = 2
    with tmpdir.as_cwd():
        generate_mask(experiments, params)

    assert all(tmpdir.join(mask).check() for mask in masks)


def test_generate_mask_with_untrusted_polygon_and_pixels(experiments_masks, tmpdir):
    experiments, masks = experiments_masks

    params = phil_scope.fetch(
        phil.parse(
            """
untrusted {
  polygon = 100 100 100 200 200 200 200 100
}
untrusted {
  pixel = 0 0
}
untrusted {
  pixel = 1 1
}"""
        )
    ).extract()

    with tmpdir.as_cwd():
        generate_mask(experiments, params)

    assert all(tmpdir.join(mask).check() for mask in masks)
    with tmpdir.join(masks[0]).open("rb") as fh:
        mask = pickle.load(fh)
    assert not mask[0][0, 0]
    assert not mask[0][1, 1]
    assert mask[0][0, 1]


def test_generate_mask_function_with_untrusted_rectangle(experiments_masks, tmpdir):
    experiments, masks = experiments_masks
    masks = [tmpdir.join(mask.replace("pixels", "pixels4")) for mask in masks]

    params = phil_scope.fetch().extract()
    params.output.mask = tmpdir.join("pixels4.mask").strpath
    params.output.experiments = tmpdir.join("masked.expt").strpath
    params.untrusted.rectangle = [100, 200, 100, 200]
    generate_mask(experiments, params)

    assert all(mask.check() for mask in masks)
    assert tmpdir.join("masked.expt").check()

    experiments = ExperimentList.from_file(tmpdir / "masked.expt")
    associated_masks = [
        imageset.external_lookup.mask.filename for imageset in experiments.imagesets()
    ]
    assert all(assoc_mask == mask for assoc_mask, mask in zip(associated_masks, masks))


def test_generate_mask_trusted_range(dials_data, tmpdir):
    # https://github.com/dials/dials/issues/978

    image_files = sorted(
        str(f) for f in dials_data("x4wide", pathlib=True).glob("*.cbf")
    )
    with tmpdir.as_cwd():
        # Import as usual
        do_import(
            ["output.experiments=no-overloads.expt"] + image_files,
            phil=import_phil_scope,
        )

        experiments = ExperimentList.from_file(tmpdir / "no-overloads.expt")
        params = phil_scope.fetch(
            phil.parse("untrusted.rectangle=100,200,100,200")
        ).extract()
        params.output.mask = "pixels1.mask"
        generate_mask(experiments, params)

        # Import with narrow trusted range to produce overloads
        do_import(
            ["trusted_range=0,100", "output.experiments=overloads.expt"] + image_files,
            phil=import_phil_scope,
        )

        experiments = ExperimentList.from_file(tmpdir / "overloads.expt")
        params = phil_scope.fetch(
            phil.parse("untrusted.rectangle=100,200,100,200")
        ).extract()
        params.output.mask = "pixels2.mask"
        generate_mask(experiments, params)

    with tmpdir.join("pixels1.mask").open("rb") as fh:
        mask1 = pickle.load(fh)
    with tmpdir.join("pixels2.mask").open("rb") as fh:
        mask2 = pickle.load(fh)

    # Overloads should not be included in the mask
    assert (mask1[0] == mask2[0]).all_eq(True)


def test_generate_whole_panel_mask(experiments_masks, tmpdir):
    experiments, masks = experiments_masks

    params = phil_scope.fetch(
        phil.parse(
            """
untrusted {
  panel = 0
}
"""
        )
    ).extract()

    with tmpdir.as_cwd():
        generate_mask(experiments, params)

    assert all(tmpdir.join(mask).check() for mask in masks)
    with tmpdir.join(masks[0]).open("rb") as fh:
        mask = pickle.load(fh)
    assert mask[0].count(False) == len(mask[0])



from __future__ import annotations

import pickle

import wx
from wx.lib.agw.floatspin import EVT_FLOATSPIN, FloatSpin

import wxtbx
from wxtbx import metallicbutton
from wxtbx.phil_controls import EVT_PHIL_CONTROL
from wxtbx.phil_controls.floatctrl import FloatCtrl as _FloatCtrl
from wxtbx.phil_controls.intctrl import IntCtrl
from wxtbx.phil_controls.strctrl import StrCtrl

import dials.util.masking


class FloatCtrl(_FloatCtrl):

    # override OnFocusLostMethod since calling event.Skip() causes bad things to
    # happen (for reasons I don't understand)
    def OnFocusLost(self, event):
        self.DoSendEvent()
        # event.Skip()


class MaskSettingsFrame(wx.MiniFrame):
    def __init__(self, *args, **kwds):
        super().__init__(*args, **kwds)
        szr = wx.BoxSizer(wx.VERTICAL)
        self.phil_params = args[0].params
        panel = MaskSettingsPanel(self)
        self.SetSizer(szr)
        szr.Add(panel, 1, wx.EXPAND)
        szr.Fit(panel)
        self.panel = panel
        self.sizer = szr
        self.Fit()
        self.Bind(wx.EVT_CLOSE, lambda evt: self.Destroy(), self)


class MaskSettingsPanel(wx.Panel):
    def __init__(self, *args, **kwds):
        super().__init__(*args, **kwds)

        self.params = args[0].phil_params

        self._pyslip = self.GetParent().GetParent().pyslip
        self.border_ctrl = None
        self.d_min_ctrl = None
        self.d_max_ctrl = None
        self.resolution_range_d_min_ctrl = None
        self.resolution_range_d_max_ctrl = None
        self.ice_rings_d_min_ctrl = None
        self.ice_rings_width_ctrl = None
        self._mode_rectangle_layer = None
        self._mode_polygon_layer = None
        self._mode_circle_layer = None
        self._rectangle_x0y0 = None
        self._rectangle_x1y1 = None
        self._mode_rectangle = False
        self._mode_polygon = False
        self._mode_polygon_points = []
        self._mode_circle = False
        self._resolution_range_d_min = 0
        self._resolution_range_d_max = 0

        self._pyslip.Bind(wx.EVT_LEFT_DOWN, self.OnLeftDown)
        self._pyslip.Bind(wx.EVT_LEFT_UP, self.OnLeftUp)
        self._pyslip.Bind(wx.EVT_MOTION, self.OnMove)
        self.draw_settings()
        self.UpdateMask()

    def __del__(self):
        if self._mode_rectangle_layer:
            self._pyslip.DeleteLayer(self._mode_rectangle_layer)
        if self._mode_polygon_layer:
            self._pyslip.DeleteLayer(self._mode_polygon_layer)
        if self._mode_circle_layer:
            self._pyslip.DeleteLayer(self._mode_circle_layer)

        self._pyslip.Unbind(wx.EVT_LEFT_DOWN, handler=self.OnLeftDown)
        self._pyslip.Unbind(wx.EVT_LEFT_UP, handler=self.OnLeftUp)
        self._pyslip.Unbind(wx.EVT_MOTION, handler=self.OnMove)

    def draw_settings(self):
        for child in self.GetChildren():
            if not isinstance(child, FloatSpin):
                # don't destroy FloatSpin controls otherwise bad things happen
                child.Destroy()

        sizer = self.GetSizer()
        if sizer is None:
            sizer = wx.BoxSizer(wx.VERTICAL)
            self.SetSizer(sizer)

        sizer.Clear()
        # Put the border, d_min, d_max in an aligned grid
        box = wx.FlexGridSizer(cols=4, hgap=0, vgap=0)

        # border control
        if self.border_ctrl is None:
            self.border_ctrl = FloatSpin(self, digits=0, name="mask_border", min_val=0)
        box.Add(
            wx.StaticText(self, label="Border"), 0, wx.ALL | wx.ALIGN_CENTER_VERTICAL, 5
        )
        box.Add(
            self.border_ctrl,
            0,
            wx.RIGHT | wx.TOP | wx.BOTTOM | wx.ALIGN_CENTER_VERTICAL,
            5,
        )
        box.Add(wx.StaticText(self, label="px"), 0, wx.ALIGN_CENTER_VERTICAL)
        self.Bind(EVT_FLOATSPIN, self.OnUpdate, self.border_ctrl)
        # Empty cell after border controls
        box.Add((0, 0))

        # d_min control
        if self.params.masking.d_min is not None:
            self.d_min = self.params.masking.d_min
        else:
            self.d_min = 0
        if self.d_min_ctrl is None:
            self.d_min_ctrl = FloatSpin(
                self,
                digits=2,
                name="d_min",
                value=self.d_min,
                min_val=0,
                increment=0.05,
            )
        txtd = wx.StaticText(self, label="d_min")
        box.Add(txtd, 0, wx.ALL | wx.ALIGN_CENTER_VERTICAL, 5)
        box.Add(
            self.d_min_ctrl,
            0,
            wx.RIGHT | wx.TOP | wx.BOTTOM | wx.ALIGN_CENTER_VERTICAL,
            5,
        )
        self.Bind(EVT_FLOATSPIN, self.OnUpdate, self.d_min_ctrl)

        # d_max control
        if self.params.masking.d_max is not None:
            self.d_max = self.params.masking.d_max
        else:
            self.d_max = 0
        if self.d_max_ctrl is None:
            self.d_max_ctrl = FloatSpin(
                self,
                digits=2,
                name="d_max",
                value=self.d_max,
                min_val=0,
                increment=0.05,
            )
        txtd = wx.StaticText(self, label="d_max")
        box.Add(txtd, 0, wx.ALL | wx.ALIGN_CENTER_VERTICAL, 5)
        box.Add(
            self.d_max_ctrl,
            0,
            wx.RIGHT | wx.TOP | wx.BOTTOM | wx.ALIGN_CENTER_VERTICAL,
            5,
        )
        self.Bind(EVT_FLOATSPIN, self.OnUpdate, self.d_max_ctrl)
        sizer.Add(box)

        # resolution rings control

        grid = wx.FlexGridSizer(
            cols=2, rows=len(self.params.masking.resolution_range) + 2, vgap=0, hgap=0
        )
        sizer.Add(grid)
        text = wx.StaticText(self, -1, "Resolution range:")
        grid.Add(text)
        # empty cell
        grid.Add(wx.StaticText(self, -1, ""), 0, wx.EXPAND)

        for range_id, (d_min, d_max) in enumerate(self.params.masking.resolution_range):
            grid.Add(wx.StaticText(self, -1, f"{d_min:.2f}-{d_max:.2f}"))
            btn = metallicbutton.MetallicButton(
                parent=self,
                label="delete",
                bmp=wxtbx.bitmaps.fetch_icon_bitmap("actions", "cancel", 16),
            )
            grid.Add(btn)
            self.Bind(
                wx.EVT_BUTTON,
                lambda evt, range_id=range_id: self.OnDeleteResolutionRange(
                    evt, range_id=range_id
                ),
                source=btn,
            )

        self.resolution_range_d_min_ctrl = FloatCtrl(
            self, value=self._resolution_range_d_min, name="resolution_range_d_min"
        )
        self.resolution_range_d_min_ctrl.SetMin(0)
        grid.Add(
            self.resolution_range_d_min_ctrl,
            0,
            wx.RIGHT | wx.TOP | wx.BOTTOM | wx.ALIGN_CENTER_VERTICAL,
            5,
        )
        self.resolution_range_d_max_ctrl = FloatCtrl(
            self, value=self._resolution_range_d_max, name="resolution_range_d_max"
        )
        self.resolution_range_d_max_ctrl.SetMin(0)
        grid.Add(
            self.resolution_range_d_max_ctrl,
            0,
            wx.RIGHT | wx.TOP | wx.BOTTOM | wx.ALIGN_CENTER_VERTICAL,
            5,
        )
        # empty cell
        # grid.Add(wx.StaticText(self, -1, ''), 0, wx.EXPAND)
        self.Bind(EVT_PHIL_CONTROL, self.OnUpdate, self.resolution_range_d_min_ctrl)
        self.Bind(EVT_PHIL_CONTROL, self.OnUpdate, self.resolution_range_d_max_ctrl)

        # ice rings control
        box = wx.BoxSizer(wx.HORIZONTAL)

        self.ice_rings_ctrl = wx.CheckBox(self, -1, "Ice rings")
        self.ice_rings_ctrl.SetValue(self.params.masking.ice_rings.filter)
        box.Add(self.ice_rings_ctrl, 0, wx.ALL | wx.ALIGN_CENTER_VERTICAL, 5)
        self.Bind(wx.EVT_CHECKBOX, self.OnUpdate, self.ice_rings_ctrl)

        # ice rings d_min control
        if self.params.masking.ice_rings.d_min is not None:
            self.ice_rings_d_min = self.params.masking.ice_rings.d_min
        else:
            self.ice_rings_d_min = 0
        if self.ice_rings_d_min_ctrl is None:
            self.ice_rings_d_min_ctrl = FloatSpin(
                self,
                digits=2,
                name="ice_rings_d_min",
                value=self.ice_rings_d_min,
                min_val=0,
                increment=0.05,
            )
        txtd = wx.StaticText(self, label="d_min")
        box.Add(txtd, 0, wx.ALL | wx.ALIGN_CENTER_VERTICAL, 5)
        box.Add(
            self.ice_rings_d_min_ctrl,
            0,
            wx.RIGHT | wx.TOP | wx.BOTTOM | wx.ALIGN_CENTER_VERTICAL,
            5,
        )
        self.Bind(EVT_FLOATSPIN, self.OnUpdate, self.ice_rings_d_min_ctrl)

        # ice rings width control
        self.ice_rings_width = self.params.masking.ice_rings.width
        if self.ice_rings_width_ctrl is None:
            self.ice_rings_width_ctrl = FloatSpin(
                self,
                digits=3,
                name="ice_rings_width",
                value=self.ice_rings_width,
                min_val=0.001,
                increment=0.001,
            )
        txtd = wx.StaticText(self, label="width")
        box.Add(txtd, 0, wx.ALL | wx.ALIGN_CENTER_VERTICAL, 5)
        box.Add(
            self.ice_rings_width_ctrl,
            0,
            wx.RIGHT | wx.TOP | wx.BOTTOM | wx.ALIGN_CENTER_VERTICAL,
            5,
        )
        self.Bind(EVT_FLOATSPIN, self.OnUpdate, self.ice_rings_width_ctrl)
        sizer.Add(box)

        untrusted_rectangles = []
        untrusted_polygons = []
        untrusted_circles = []

        # map index in self.params.masking.untrusted to index in above arrays
        self._rectangle_to_untrusted_id = []
        self._polygon_to_untrusted_id = []
        self._circle_to_untrusted_id = []

        for i, untrusted in enumerate(self.params.masking.untrusted):
            if untrusted.rectangle is not None:
                untrusted_rectangles.append((untrusted.panel, untrusted.rectangle))
                self._rectangle_to_untrusted_id.append(i)
            elif untrusted.polygon is not None:
                untrusted_polygons.append((untrusted.panel, untrusted.polygon))
                self._polygon_to_untrusted_id.append(i)
            elif untrusted.circle is not None:
                untrusted_circles.append((untrusted.panel, untrusted.circle))
                self._circle_to_untrusted_id.append(i)

        # untrusted rectangles
        grid = wx.FlexGridSizer(
            cols=3, rows=len(untrusted_rectangles) + 2, vgap=0, hgap=0
        )
        sizer.Add(grid)
        text = wx.StaticText(self, -1, "Panel:")
        text.GetFont().SetWeight(wx.BOLD)
        grid.Add(text)
        text = wx.StaticText(self, -1, "Rectangle (x0, x1, y0, y1):")
        text.GetFont().SetWeight(wx.BOLD)
        grid.Add(text)
        # empty cell
        grid.Add(wx.StaticText(self, -1, ""), 0, wx.EXPAND)

        for rect_id, (panel, rectangle) in enumerate(untrusted_rectangles):
            grid.Add(wx.StaticText(self, -1, "%i" % (panel)))
            grid.Add(wx.StaticText(self, -1, "%i %i %i %i" % tuple(rectangle)))
            btn = metallicbutton.MetallicButton(
                parent=self,
                label="delete",
                bmp=wxtbx.bitmaps.fetch_icon_bitmap("actions", "cancel", 16),
            )
            grid.Add(btn)
            untrusted_id = self._rectangle_to_untrusted_id[rect_id]
            self.Bind(
                wx.EVT_BUTTON,
                lambda evt, untrusted_id=untrusted_id: self.OnDeleteUntrustedRegion(
                    evt, untrusted_id=untrusted_id
                ),
                source=btn,
            )

        self.untrusted_rectangle_panel_ctrl = IntCtrl(
            self, value=0, name="untrusted_rectangle_panel"
        )
        grid.Add(self.untrusted_rectangle_panel_ctrl, 0, wx.ALL, 5)
        self.untrusted_rectangle_ctrl = StrCtrl(
            self, value="", name="untrusted_rectangle"
        )
        grid.Add(self.untrusted_rectangle_ctrl, 0, wx.ALL, 5)
        # empty cell
        grid.Add(wx.StaticText(self, -1, ""), 0, wx.EXPAND)
        self.Bind(EVT_PHIL_CONTROL, self.OnUpdate, self.untrusted_rectangle_ctrl)

        # untrusted polygons
        grid = wx.FlexGridSizer(
            cols=3, rows=len(untrusted_polygons) + 2, vgap=0, hgap=0
        )
        sizer.Add(grid)
        text = wx.StaticText(self, -1, "Panel:")
        text.GetFont().SetWeight(wx.BOLD)
        grid.Add(text)
        text = wx.StaticText(self, -1, "Polygons (x1, y1, ..., xn, yn):")
        text.GetFont().SetWeight(wx.BOLD)
        grid.Add(text)
        # empty cell
        grid.Add(wx.StaticText(self, -1, ""), 0, wx.EXPAND)

        for polygon_id, (panel, polygon) in enumerate(untrusted_polygons):
            grid.Add(
                StrCtrl(self, value="%i" % (panel), style=wx.TE_READONLY), 0, wx.ALL, 5
            )
            grid.Add(
                StrCtrl(
                    self,
                    value=" ".join(["%i"] * len(polygon)) % tuple(polygon),
                    style=wx.TE_READONLY,
                ),
                0,
                wx.ALL,
                5,
            )
            btn = metallicbutton.MetallicButton(
                parent=self,
                label="delete",
                bmp=wxtbx.bitmaps.fetch_icon_bitmap("actions", "cancel", 16),
            )
            grid.Add(btn)
            untrusted_id = self._polygon_to_untrusted_id[polygon_id]
            self.Bind(
                wx.EVT_BUTTON,
                lambda evt, untrusted_id=untrusted_id: self.OnDeleteUntrustedRegion(
                    evt, untrusted_id=untrusted_id
                ),
                source=btn,
            )

        self.untrusted_polygon_panel_ctrl = IntCtrl(
            self, value=0, name="untrusted_polygon_panel"
        )
        grid.Add(self.untrusted_polygon_panel_ctrl, 0, wx.ALL, 5)
        self.untrusted_polygon_ctrl = StrCtrl(self, value="", name="untrusted_polygon")
        grid.Add(self.untrusted_polygon_ctrl, 0, wx.ALL, 5)
        # empty cell
        grid.Add(wx.StaticText(self, -1, ""), 0, wx.EXPAND)
        self.Bind(EVT_PHIL_CONTROL, self.OnUpdate, self.untrusted_polygon_ctrl)

        # untrusted circles
        grid = wx.FlexGridSizer(cols=3, rows=len(untrusted_circles) + 2, vgap=0, hgap=0)
        sizer.Add(grid)
        text = wx.StaticText(self, -1, "Panel:")
        text.GetFont().SetWeight(wx.BOLD)
        grid.Add(text)
        text = wx.StaticText(self, -1, "Circle (x, y, r):")
        text.GetFont().SetWeight(wx.BOLD)
        grid.Add(text)
        # empty cell
        grid.Add(wx.StaticText(self, -1, ""), 0, wx.EXPAND)

        for circle_id, (panel, circle) in enumerate(untrusted_circles):
            grid.Add(wx.StaticText(self, -1, "%i" % (panel)))
            grid.Add(wx.StaticText(self, -1, "%i %i %i" % tuple(circle)))
            btn = metallicbutton.MetallicButton(
                parent=self,
                label="delete",
                bmp=wxtbx.bitmaps.fetch_icon_bitmap("actions", "cancel", 16),
            )
            grid.Add(btn)
            untrusted_id = self._circle_to_untrusted_id[circle_id]
            self.Bind(
                wx.EVT_BUTTON,
                lambda evt, untrusted_id=untrusted_id: self.OnDeleteUntrustedRegion(
                    evt, untrusted_id=untrusted_id
                ),
                source=btn,
            )

        self.untrusted_circle_panel_ctrl = IntCtrl(
            self, value=0, name="untrusted_circle_panel"
        )
        grid.Add(self.untrusted_circle_panel_ctrl, 0, wx.ALL, 5)
        self.untrusted_circle_ctrl = StrCtrl(self, value="", name="untrusted_circle")
        grid.Add(self.untrusted_circle_ctrl, 0, wx.ALL, 5)
        # empty cell
        grid.Add(wx.StaticText(self, -1, ""), 0, wx.EXPAND)
        self.Bind(EVT_PHIL_CONTROL, self.OnUpdate, self.untrusted_circle_ctrl)

        # Draw rectangle/circle mode buttons
        grid = wx.FlexGridSizer(cols=4, rows=1, vgap=0, hgap=0)
        sizer.Add(grid)

        grid.Add(wx.StaticText(self, label="Mode:"))
        self.mode_rectangle_button = wx.ToggleButton(self, -1, "Rectangle")
        self.mode_rectangle_button.SetValue(self._mode_rectangle)
        grid.Add(self.mode_rectangle_button, 0, wx.ALL | wx.ALIGN_CENTER_VERTICAL, 5)
        self.Bind(wx.EVT_TOGGLEBUTTON, self.OnUpdate, self.mode_rectangle_button)

        self.mode_circle_button = wx.ToggleButton(self, -1, "Circle")
        self.mode_circle_button.SetValue(self._mode_circle)
        grid.Add(self.mode_circle_button, 0, wx.ALL | wx.ALIGN_CENTER_VERTICAL, 5)
        self.Bind(wx.EVT_TOGGLEBUTTON, self.OnUpdate, self.mode_circle_button)

        self.mode_polygon_button = wx.ToggleButton(self, -1, "Polygon")
        self.mode_polygon_button.SetValue(self._mode_polygon)
        grid.Add(self.mode_polygon_button, 0, wx.ALL | wx.ALIGN_CENTER_VERTICAL, 5)
        self.Bind(wx.EVT_TOGGLEBUTTON, self.OnUpdate, self.mode_polygon_button)

        # show/save mask controls
        grid = wx.FlexGridSizer(cols=3, rows=2, vgap=0, hgap=0)
        sizer.Add(grid)

        self.show_mask_ctrl = wx.CheckBox(self, -1, "Show mask")
        self.show_mask_ctrl.SetValue(self.params.show_mask)
        grid.Add(self.show_mask_ctrl, 0, wx.ALL | wx.ALIGN_CENTER_VERTICAL, 5)
        self.Bind(wx.EVT_CHECKBOX, self.OnUpdate, self.show_mask_ctrl)

        self.save_mask_button = wx.Button(self, -1, "Save mask")
        grid.Add(self.save_mask_button, 0, wx.ALL | wx.ALIGN_CENTER_VERTICAL, 5)
        self.Bind(wx.EVT_BUTTON, self.OnSaveMask, self.save_mask_button)

        self.save_mask_txt_ctrl = StrCtrl(
            self, value=self.params.output.mask, name="mask_pickle"
        )
        grid.Add(self.save_mask_txt_ctrl, 0, wx.ALL, 5)
        self.Bind(EVT_PHIL_CONTROL, self.OnUpdate, self.save_mask_txt_ctrl)

        # empty cell
        grid.Add(wx.StaticText(self, -1, ""), 0, wx.EXPAND)

        self.save_params_button = wx.Button(self, -1, "Save")
        grid.Add(self.save_params_button, 0, wx.ALL | wx.ALIGN_CENTER_VERTICAL, 5)
        self.Bind(wx.EVT_BUTTON, self.OnSaveMaskParams, self.save_params_button)

        self.save_params_txt_ctrl = StrCtrl(
            self, value=self.params.output.mask_params, name="mask_phil"
        )
        grid.Add(self.save_params_txt_ctrl, 0, wx.ALL, 5)
        self.Bind(EVT_PHIL_CONTROL, self.OnUpdate, self.save_params_txt_ctrl)

        sizer.Layout()
        sizer.Fit(self)

    def OnDeleteUntrustedRegion(self, event, untrusted_id):
        if untrusted_id is not None:
            del self.params.masking.untrusted[untrusted_id]
        self.OnUpdate(event)

    def OnDeleteResolutionRange(self, event, range_id):
        if range_id is not None:
            del self.params.masking.resolution_range[range_id]
        self.OnUpdate(event)

    def OnUpdate(self, event):
        image_viewer_frame = self.GetParent().GetParent()

        self.params.show_mask = self.show_mask_ctrl.GetValue()
        # untidy
        image_viewer_frame.settings.show_mask = self.params.show_mask
        image_viewer_frame.params.show_mask = self.params.show_mask
        image_viewer_frame.settings_frame.panel.show_mask.SetValue(
            self.params.show_mask
        )

        self.params.output.mask = self.save_mask_txt_ctrl.GetValue()
        self.params.output.mask_params = self.save_params_txt_ctrl.GetValue()

        if self._mode_polygon:
            self.AddUntrustedPolygon(self._mode_polygon_points)
            self._mode_polygon_points = []
            self._pyslip.DeleteLayer(self._mode_polygon_layer)
            self._mode_polygon_layer = None

        if self.mode_rectangle_button.GetValue():
            if not self._mode_rectangle:
                # mode wasn't set but button has been pressed, so set mode
                self._mode_rectangle = True
                # set other modes and buttons to False
                self._mode_circle = False
                self.mode_circle_button.SetValue(False)
                self._mode_polygon = False
                self.mode_polygon_button.SetValue(False)

        if self.mode_circle_button.GetValue():
            if not self._mode_circle:
                # mode wasn't set but button has been pressed, so set mode
                self._mode_circle = True
                # set other modes and buttons to False
                self._mode_rectangle = False
                self.mode_rectangle_button.SetValue(False)
                self._mode_polygon = False
                self.mode_polygon_button.SetValue(False)

        if self.mode_polygon_button.GetValue():
            if not self._mode_polygon:
                # mode wasn't set but button has been pressed, so set mode
                self._mode_polygon = True
                # set other modes and buttons to False
                self._mode_rectangle = False
                self.mode_rectangle_button.SetValue(False)
                self._mode_circle = False
                self.mode_circle_button.SetValue(False)

        if not (
            self.mode_circle_button.GetValue()
            or self.mode_rectangle_button.GetValue()
            or self.mode_polygon_button.GetValue()
        ):
            self._mode_circle = False
            self._mode_rectangle = False
            self._mode_polygon = False

        if self.d_min_ctrl.GetValue() > 0:
            self.params.masking.d_min = self.d_min_ctrl.GetValue()
        else:
            self.params.masking.d_min = None
        if self.d_max_ctrl.GetValue() > 0:
            self.params.masking.d_max = self.d_max_ctrl.GetValue()
        else:
            self.params.masking.d_max = None
        self.params.masking.border = int(self.border_ctrl.GetValue())

        if self.ice_rings_d_min_ctrl.GetValue() > 0:
            self.params.masking.ice_rings.d_min = self.ice_rings_d_min_ctrl.GetValue()
        else:
            self.params.masking.ice_rings.d_min = None
        if self.ice_rings_width_ctrl.GetValue() > 0:
            self.params.masking.ice_rings.width = self.ice_rings_width_ctrl.GetValue()
        self.params.masking.ice_rings.filter = self.ice_rings_ctrl.GetValue()

        self._resolution_range_d_min = float(
            self.resolution_range_d_min_ctrl.GetValue()
        )
        self._resolution_range_d_max = float(
            self.resolution_range_d_max_ctrl.GetValue()
        )

        if self._resolution_range_d_min > 0 and self._resolution_range_d_max > 0:

            self.params.masking.resolution_range.append(
                (self._resolution_range_d_min, self._resolution_range_d_max)
            )
            self._resolution_range_d_min = 0
            self._resolution_range_d_max = 0

        from dials.util import masking

        untrusted_rectangle = self.untrusted_rectangle_ctrl.GetValue().strip()
        if len(untrusted_rectangle.strip()) > 0:
            rectangle = untrusted_rectangle.strip().replace(",", " ").split(" ")
            try:
                rectangle = [int(s) for s in rectangle]
                assert len(rectangle) == 4
                panel = int(self.untrusted_rectangle_panel_ctrl.GetValue())
            except Exception:
                pass
            else:
                untrusted = masking.phil_scope.extract().untrusted[0]
                untrusted.panel = panel
                untrusted.rectangle = rectangle
                self.params.masking.untrusted.append(untrusted)

        untrusted_polygon = self.untrusted_polygon_ctrl.GetValue().strip()
        if len(untrusted_polygon.strip()) > 0:
            polygon = untrusted_polygon.strip().replace(",", " ").split(" ")
            try:
                polygon = [int(s) for s in polygon]
                assert len(polygon) % 2 == 0
                assert len(polygon) // 2 > 3
                panel = int(self.untrusted_polygon_panel_ctrl.GetValue())
            except Exception:
                pass
            else:
                untrusted = masking.phil_scope.extract().untrusted[0]
                untrusted.panel = panel
                untrusted.polygon = polygon
                self.params.masking.untrusted.append(untrusted)

        untrusted_circle = self.untrusted_circle_ctrl.GetValue().strip()
        if len(untrusted_circle.strip()) > 0:
            circle = untrusted_circle.strip().replace(",", " ").split(" ")
            try:
                circle = [int(s) for s in circle]
                assert len(circle) == 3
                panel = int(self.untrusted_circle_panel_ctrl.GetValue())
            except Exception:
                pass
            else:
                untrusted = masking.phil_scope.extract().untrusted[0]
                untrusted.panel = panel
                untrusted.circle = circle
                self.params.masking.untrusted.append(untrusted)

        # Refresh the UI *after* the event is finished
        wx.CallAfter(self.draw_settings)

        self.UpdateMask()

        # Force re-drawing of mask
        image_viewer_frame.OnChooseImage(event)

    def OnSaveMask(self, event):
        self.OnUpdate(event)
        self.UpdateMask()
        image_viewer_frame = self.GetParent().GetParent()

        m1 = image_viewer_frame.mask_input
        m2 = image_viewer_frame.mask_image_viewer

        if m1 is not None and m2 is not None:
            mask = []
            for p1, p2 in zip(m1, m2):
                mask.append(p2 & p1)
            mask = tuple(mask)
        elif m1 is not None:
            mask = m1
        elif m2 is not None:
            mask = m2
        else:
            return

        print(f"Writing mask to {self.params.output.mask}")
        with open(self.params.output.mask, "wb") as fh:
            pickle.dump(mask, fh, protocol=pickle.HIGHEST_PROTOCOL)

    def OnSaveMaskParams(self, event):
        self.OnUpdate(event)
        file_name = self.params.output.mask_params
        with open(file_name, "w") as f:
            print(f"Saving parameters to {file_name}")
            dials.util.masking.phil_scope.fetch_diff(
                dials.util.masking.phil_scope.format(self.params.masking)
            ).show(f)

    def UpdateMask(self):
        image_viewer_frame = self.GetParent().GetParent()
        mask = dials.util.masking.generate_mask(
            image_viewer_frame.imagesets[0], self.params.masking
        )
        image_viewer_frame.mask_image_viewer = mask
        image_viewer_frame.update_settings(layout=False)

    def OnLeftDown(self, event):
        if not event.ShiftDown():
            click_posn = event.GetPosition()
            if self._mode_rectangle:
                self._rectangle_x0y0 = click_posn
                self._rectangle_x1y1 = None
                return
            elif self._mode_circle:
                self._circle_xy = click_posn
                self._circle_radius = None
                return
            elif self._mode_polygon:
                self._mode_polygon_points.append(click_posn)
                self.DrawPolygon(self._mode_polygon_points)
        event.Skip()

    def OnLeftUp(self, event):
        if not event.ShiftDown():
            click_posn = event.GetPosition()

            if self._mode_rectangle and self._rectangle_x0y0 is not None:
                self._rectangle_x1y1 = click_posn
                x0, y0 = self._rectangle_x0y0
                x1, y1 = self._rectangle_x1y1
                self.AddUntrustedRectangle(x0, y0, x1, y1)
                self._pyslip.DeleteLayer(self._mode_rectangle_layer)
                self._mode_rectangle_layer = None
                self.mode_rectangle_button.SetValue(False)
                self.OnUpdate(event)
                return

            elif self._mode_circle and self._circle_xy is not None:
                xc, yc = self._circle_xy
                xedge, yedge = click_posn
                self.DrawCircle(xc, yc, xedge, yedge)
                try:
                    self.AddUntrustedCircle(xc, yc, xedge, yedge)
                except Exception as e:
                    print(e)
                finally:
                    self._pyslip.DeleteLayer(self._mode_circle_layer)
                    self._mode_circle_layer = None
                    self.mode_circle_button.SetValue(False)
                    self.OnUpdate(event)
                    return

        event.Skip()

    def OnMove(self, event):
        if event.Dragging() and event.LeftIsDown() and not event.ShiftDown():
            if self._mode_rectangle:
                if self._rectangle_x0y0 is not None:
                    x0, y0 = self._rectangle_x0y0
                    x1, y1 = event.GetPosition()
                    self.DrawRectangle(x0, y0, x1, y1)
                    return

            elif self._mode_circle:
                if self._circle_xy is not None:
                    xc, yc = self._circle_xy
                    xedge, yedge = event.GetPosition()
                    self.DrawCircle(xc, yc, xedge, yedge)
                    return
        event.Skip()

    def DrawRectangle(self, x0, y0, x1, y1):
        if self._mode_rectangle_layer:
            self._pyslip.DeleteLayer(self._mode_rectangle_layer)
            self._mode_rectangle_layer = None

        polygon = [(x0, y0), (x1, y0), (x1, y1), (x0, y1), (x0, y0)]
        d = {}
        polygon_data = []
        points = [self._pyslip.ConvertView2Geo(p) for p in polygon]
        for i in range(len(points) - 1):
            polygon_data.append(((points[i], points[i + 1]), d))

        self._mode_rectangle_layer = self._pyslip.AddPolygonLayer(
            polygon_data,
            map_rel=True,
            color="#00ffff",
            radius=5,
            visible=True,
            name="<mode_rectangle_layer>",
        )

    def DrawCircle(self, xc, yc, xedge, yedge):
        if self._mode_circle_layer:
            self._pyslip.DeleteLayer(self._mode_circle_layer)
            self._mode_circle_layer = None

        xc, yc = self._pyslip.ConvertView2Geo((xc, yc))
        xedge, yedge = self._pyslip.ConvertView2Geo((xedge, yedge))

        from scitbx import matrix

        center = matrix.col((xc, yc))
        edge = matrix.col((xedge, yedge))
        r = (center - edge).length()
        if r == 0:
            return

        e1 = matrix.col((1, 0))
        e2 = matrix.col((0, 1))
        circle_data = (
            center + r * (e1 + e2),
            center + r * (e1 - e2),
            center + r * (-e1 - e2),
            center + r * (-e1 + e2),
            center + r * (e1 + e2),
        )

        self._mode_circle_layer = self._pyslip.AddEllipseLayer(
            circle_data,
            map_rel=True,
            color="#00ffff",
            radius=5,
            visible=True,
            # show_levels=[3,4],
            name="<mode_circle_layer>",
        )

    def DrawPolygon(self, vertices):

        if self._mode_polygon_layer:
            self._pyslip.DeleteLayer(self._mode_polygon_layer)
            self._mode_polygon_layer = None

        polygon_data = []
        d = {}

        for i in range(len(vertices) - 1):
            polygon_data.append(
                (
                    (
                        self._pyslip.ConvertView2Geo(vertices[i]),
                        self._pyslip.ConvertView2Geo(vertices[i + 1]),
                    ),
                    d,
                )
            )

        if polygon_data:
            self._mode_polygon_layer = self._pyslip.AddPolygonLayer(
                polygon_data,
                map_rel=True,
                color="#00ffff",
                radius=5,
                visible=True,
                name="<boxsel_pt_layer>",
            )

    def AddUntrustedPolygon(self, vertices):
        if len(vertices) < 4:
            return
        vertices.append(vertices[0])
        vertices = [self._pyslip.ConvertView2Geo(v) for v in vertices]
        vertices = [
            self._pyslip.tiles.map_relative_to_picture_fast_slow(*v) for v in vertices
        ]

        point_ = []
        panel_id = None
        for p in vertices:
            p1, p0, p_id = self._pyslip.tiles.flex_image.picture_to_readout(p[1], p[0])
            assert p_id >= 0, "Point must be within a panel"
            if panel_id is not None:
                assert (
                    panel_id == p_id
                ), "All points must be contained within a single panel"
            panel_id = p_id
            point_.append((p0, p1))
        vertices = point_

        from libtbx.utils import flat_list

        from dials.util import masking

        region = masking.phil_scope.extract().untrusted[0]
        points = flat_list(vertices)
        region.polygon = [int(p) for p in points]
        region.panel = panel_id

        self.params.masking.untrusted.append(region)

    def AddUntrustedRectangle(self, x0, y0, x1, y1):
        x0, y0 = self._pyslip.ConvertView2Geo((x0, y0))
        x1, y1 = self._pyslip.ConvertView2Geo((x1, y1))

        if x0 == x1 or y0 == y1:
            return

        points = [(x0, y0), (x1, y1)]
        points = [
            self._pyslip.tiles.map_relative_to_picture_fast_slow(*p) for p in points
        ]

        detector = self._pyslip.tiles.raw_image.get_detector()
        point_ = []
        panel_id = None
        for p in points:
            p1, p0, p_id = self._pyslip.tiles.flex_image.picture_to_readout(p[1], p[0])
            assert p_id >= 0, "Point must be within a panel"
            if panel_id is not None:
                assert (
                    panel_id == p_id
                ), "All points must be contained within a single panel"
            panel_id = int(p_id)
            point_.append((p0, p1))
        points = point_

        (x0, y0), (x1, y1) = points

        if x0 > x1:
            x1, x0 = x0, x1
        if y0 > y1:
            y1, y0 = y0, y1

        panel = detector[panel_id]
        if (
            x1 < 0
            or y1 < 0
            or x0 > panel.get_image_size()[0]
            or y0 > panel.get_image_size()[1]
        ):
            return

        x0 = max(0, x0)
        y0 = max(0, y0)
        x1 = min(panel.get_image_size()[0], x1)
        y1 = min(panel.get_image_size()[1], y1)

        from dials.util import masking

        region = masking.phil_scope.extract().untrusted[0]
        region.rectangle = [int(x0), int(x1), int(y0), int(y1)]
        region.panel = panel_id

        self.params.masking.untrusted.append(region)

    def AddUntrustedCircle(self, xc, yc, xedge, yedge):

        points = [(xc, yc), (xedge, yedge)]

        points = [self._pyslip.ConvertView2Geo(p) for p in points]
        points = [
            self._pyslip.tiles.map_relative_to_picture_fast_slow(*p) for p in points
        ]

        points_ = []
        panel_id = None
        for p in points:
            p1, p0, p_id = self._pyslip.tiles.flex_image.picture_to_readout(p[1], p[0])
            assert p_id >= 0, "Point must be within a panel"
            if panel_id is not None:
                assert (
                    panel_id == p_id
                ), "All points must be contained within a single panel"
            panel_id = p_id
            points_.append((p0, p1))
        points = points_

        (xc, yc), (xedge, yedge) = points

        from scitbx import matrix

        center = matrix.col((xc, yc))
        edge = matrix.col((xedge, yedge))
        r = (center - edge).length()
        if r == 0:
            return

        from dials.util import masking

        region = masking.phil_scope.extract().untrusted[0]
        region.circle = [int(xc), int(yc), int(r)]
        region.panel = panel_id

        self.params.masking.untrusted.append(region)



from __future__ import annotations

from pathlib import Path

import procrunner


def test(dials_regression, tmp_path):
    experiments_path = Path(dials_regression) / "misc_test_data" / "i04-indexed.json"
    pickle_path = Path(dials_regression) / "misc_test_data" / "i04-indexed.pickle"

    assert experiments_path.is_file()
    assert pickle_path.is_file()

    result = procrunner.run(
        [
            "dials.check_indexing_symmetry",
            experiments_path,
            pickle_path,
            "d_min=4",
            "d_max=10",
            "symop_threshold=0.6",
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr



from __future__ import annotations


def export_text(integrated_data):
    """Export contents of a dials reflection table as text."""

    h, k, l = zip(*integrated_data["miller_index"])

    # FIXME Currently outputting either summation or profile fitting. Should do
    # both?
    if "intensity.prf" in integrated_data:
        i = integrated_data["intensity.prf.value"]
        v = integrated_data["intensity.prf.variance"]
    else:
        i = integrated_data["intensity.sum.value"]
        v = integrated_data["intensity.sum.variance"]
    lp = integrated_data["lp"]
    i *= lp
    v *= lp

    for _h, _k, _l, _i, _v in zip(h, k, l, i, v):
        print("%4d %4d %4d %f %f" % (_h, _k, _l, _i, _v))



"""
A simple test of refinement using two crystals.
"""

from __future__ import annotations

from math import pi

from cctbx.sgtbx import space_group, space_group_symbols
from cctbx.uctbx import unit_cell
from dxtbx.model import ScanFactory
from dxtbx.model.experiment_list import Experiment, ExperimentList
from libtbx.phil import parse
from libtbx.test_utils import approx_equal
from rstbx.symmetry.constraints.parameter_reduction import symmetrize_reduce_enlarge
from scitbx import matrix
from scitbx.array_family import flex

from dials.algorithms.refinement.parameterisation.beam_parameters import (
    BeamParameterisation,
)
from dials.algorithms.refinement.parameterisation.crystal_parameters import (
    CrystalOrientationParameterisation,
    CrystalUnitCellParameterisation,
)
from dials.algorithms.refinement.parameterisation.detector_parameters import (
    DetectorParameterisationSinglePanel,
)
from dials.algorithms.refinement.prediction.managed_predictors import (
    ScansExperimentsPredictor,
    ScansRayPredictor,
)
from dials.algorithms.refinement.refiner import RefinerFactory, phil_scope
from dials.algorithms.spot_prediction import IndexGenerator, ray_intersection

from . import geometry_phil, minimiser_phil, setup_geometry


def test(args=[]):

    #############################
    # Setup experimental models #
    #############################

    master_phil = parse(f"{geometry_phil}\n{minimiser_phil}")

    models = setup_geometry.Extract(
        master_phil,
        cmdline_args=args,
        local_overrides="geometry.parameters.random_seed = 1",
    )

    crystal1 = models.crystal

    models = setup_geometry.Extract(
        master_phil,
        cmdline_args=args,
        local_overrides="geometry.parameters.random_seed = 2",
    )

    mydetector = models.detector
    mygonio = models.goniometer
    crystal2 = models.crystal
    mybeam = models.beam

    # Build a mock scan for an 18 degree sequence
    sf = ScanFactory()
    myscan = sf.make_scan(
        image_range=(1, 180),
        exposure_times=0.1,
        oscillation=(0, 0.1),
        epochs=list(range(180)),
        deg=True,
    )
    sequence_range = myscan.get_oscillation_range(deg=False)
    im_width = myscan.get_oscillation(deg=False)[1]
    assert sequence_range == (0.0, pi / 10)
    assert approx_equal(im_width, 0.1 * pi / 180.0)

    # Build an experiment list
    experiments = ExperimentList()
    experiments.append(
        Experiment(
            beam=mybeam,
            detector=mydetector,
            goniometer=mygonio,
            scan=myscan,
            crystal=crystal1,
            imageset=None,
        )
    )
    experiments.append(
        Experiment(
            beam=mybeam,
            detector=mydetector,
            goniometer=mygonio,
            scan=myscan,
            crystal=crystal2,
            imageset=None,
        )
    )

    assert len(experiments.detectors()) == 1

    ##########################################################
    # Parameterise the models (only for perturbing geometry) #
    ##########################################################

    det_param = DetectorParameterisationSinglePanel(mydetector)
    s0_param = BeamParameterisation(mybeam, mygonio)
    xl1o_param = CrystalOrientationParameterisation(crystal1)
    xl1uc_param = CrystalUnitCellParameterisation(crystal1)
    xl2o_param = CrystalOrientationParameterisation(crystal2)
    xl2uc_param = CrystalUnitCellParameterisation(crystal2)

    # Fix beam to the X-Z plane (imgCIF geometry), fix wavelength
    s0_param.set_fixed([True, False, True])

    ################################
    # Apply known parameter shifts #
    ################################

    # shift detector by 1.0 mm each translation and 2 mrad each rotation
    det_p_vals = det_param.get_param_vals()
    p_vals = [a + b for a, b in zip(det_p_vals, [1.0, 1.0, 1.0, 2.0, 2.0, 2.0])]
    det_param.set_param_vals(p_vals)

    # shift beam by 2 mrad in free axis
    s0_p_vals = s0_param.get_param_vals()
    p_vals = list(s0_p_vals)

    p_vals[0] += 2.0
    s0_param.set_param_vals(p_vals)

    # rotate crystal a bit (=2 mrad each rotation)
    xlo_p_vals = []
    for xlo in (xl1o_param, xl2o_param):
        p_vals = xlo.get_param_vals()
        xlo_p_vals.append(p_vals)
        new_p_vals = [a + b for a, b in zip(p_vals, [2.0, 2.0, 2.0])]
        xlo.set_param_vals(new_p_vals)

    # change unit cell a bit (=0.1 Angstrom length upsets, 0.1 degree of
    # gamma angle)
    xluc_p_vals = []
    for xluc, xl in ((xl1uc_param, crystal1), ((xl2uc_param, crystal2))):
        p_vals = xluc.get_param_vals()
        xluc_p_vals.append(p_vals)
        cell_params = xl.get_unit_cell().parameters()
        cell_params = [
            a + b for a, b in zip(cell_params, [0.1, 0.1, 0.1, 0.0, 0.0, 0.1])
        ]
        new_uc = unit_cell(cell_params)
        newB = matrix.sqr(new_uc.fractionalization_matrix()).transpose()
        S = symmetrize_reduce_enlarge(xl.get_space_group())
        S.set_orientation(orientation=newB)
        X = tuple([e * 1.0e5 for e in S.forward_independent_parameters()])
        xluc.set_param_vals(X)

    #############################
    # Generate some reflections #
    #############################

    # All indices in a 2.5 Angstrom sphere for crystal1
    resolution = 2.5
    index_generator = IndexGenerator(
        crystal1.get_unit_cell(),
        space_group(space_group_symbols(1).hall()).type(),
        resolution,
    )
    indices1 = index_generator.to_array()

    # All indices in a 2.5 Angstrom sphere for crystal2
    resolution = 2.5
    index_generator = IndexGenerator(
        crystal2.get_unit_cell(),
        space_group(space_group_symbols(1).hall()).type(),
        resolution,
    )
    indices2 = index_generator.to_array()

    # Predict rays within the sequence range. Set experiment IDs
    ray_predictor = ScansRayPredictor(experiments, sequence_range)
    obs_refs1 = ray_predictor(indices1, experiment_id=0)
    obs_refs1["id"] = flex.int(len(obs_refs1), 0)
    obs_refs2 = ray_predictor(indices2, experiment_id=1)
    obs_refs2["id"] = flex.int(len(obs_refs2), 1)

    # Take only those rays that intersect the detector
    intersects = ray_intersection(mydetector, obs_refs1)
    obs_refs1 = obs_refs1.select(intersects)
    intersects = ray_intersection(mydetector, obs_refs2)
    obs_refs2 = obs_refs2.select(intersects)

    # Make a reflection predictor and re-predict for all these reflections. The
    # result is the same, but we gain also the flags and xyzcal.px columns
    ref_predictor = ScansExperimentsPredictor(experiments)
    obs_refs1 = ref_predictor(obs_refs1)
    obs_refs2 = ref_predictor(obs_refs2)

    # Set 'observed' centroids from the predicted ones
    obs_refs1["xyzobs.mm.value"] = obs_refs1["xyzcal.mm"]
    obs_refs2["xyzobs.mm.value"] = obs_refs2["xyzcal.mm"]

    # Invent some variances for the centroid positions of the simulated data
    im_width = 0.1 * pi / 18.0
    px_size = mydetector[0].get_pixel_size()
    var_x = flex.double(len(obs_refs1), (px_size[0] / 2.0) ** 2)
    var_y = flex.double(len(obs_refs1), (px_size[1] / 2.0) ** 2)
    var_phi = flex.double(len(obs_refs1), (im_width / 2.0) ** 2)
    obs_refs1["xyzobs.mm.variance"] = flex.vec3_double(var_x, var_y, var_phi)
    var_x = flex.double(len(obs_refs2), (px_size[0] / 2.0) ** 2)
    var_y = flex.double(len(obs_refs2), (px_size[1] / 2.0) ** 2)
    var_phi = flex.double(len(obs_refs2), (im_width / 2.0) ** 2)
    obs_refs2["xyzobs.mm.variance"] = flex.vec3_double(var_x, var_y, var_phi)

    # concatenate reflection lists
    obs_refs1.extend(obs_refs2)
    obs_refs = obs_refs1

    ###############################
    # Undo known parameter shifts #
    ###############################

    s0_param.set_param_vals(s0_p_vals)
    det_param.set_param_vals(det_p_vals)
    xl1o_param.set_param_vals(xlo_p_vals[0])
    xl2o_param.set_param_vals(xlo_p_vals[1])
    xl1uc_param.set_param_vals(xluc_p_vals[0])
    xl2uc_param.set_param_vals(xluc_p_vals[1])

    # scan static first
    params = phil_scope.fetch(source=parse("")).extract()
    refiner = RefinerFactory.from_parameters_data_experiments(
        params, obs_refs, experiments
    )
    refiner.run()

    # scan varying
    params.refinement.parameterisation.scan_varying = True
    refiner = RefinerFactory.from_parameters_data_experiments(
        params, obs_refs, experiments
    )
    refiner.run()

    # Ensure all models have scan-varying state set
    # (https://github.com/dials/dials/issues/798)
    refined_experiments = refiner.get_experiments()
    sp = [xl.get_num_scan_points() for xl in refined_experiments.crystals()]

    assert sp.count(181) == 2



"""
Tests for the dials.report.plots module.
"""

from __future__ import annotations

import itertools
import random
from unittest import mock

import pytest

from cctbx import crystal, miller, uctbx
from cctbx.array_family import flex
from iotbx.merging_statistics import dataset_statistics

from dials.report.plots import (
    AnomalousPlotter,
    IntensityStatisticsPlots,
    ResolutionPlotsAndStats,
    i_over_sig_i_vs_batch_plot,
    i_over_sig_i_vs_i_plot,
    scale_rmerge_vs_batch_plot,
)
from dials.util.batch_handling import batch_manager


@pytest.fixture
def iobs():
    """Generate a miller array for testing plotters."""
    ms = miller.build_set(
        crystal_symmetry=crystal.symmetry(
            space_group_symbol="P1", unit_cell=(6, 6, 6, 90, 90, 90)
        ),
        anomalous_flag=True,
        d_min=1.0,
    )
    data = flex.double(float(random.randrange(0, 100)) for _ in range(ms.size()))
    iobs = miller.array(ms, data, sigmas=data)
    iobs.change_symmetry(space_group_symbol="P222", merge_non_unique=False)
    iobs.set_info(miller.array_info(source="DIALS", source_type="reflection_tables"))
    iobs.set_observation_type_xray_intensity()
    return iobs


def test_AnomalousPlotter():

    "Make a larger array to allow all plots to be made"
    cs = crystal.symmetry(space_group_symbol="P1", unit_cell=(6, 6, 6, 90, 90, 90))
    ms = miller.build_set(cs, anomalous_flag=True, d_min=1.0)
    indices = ms.indices()
    new_indices = flex.miller_index(list(indices) * 10)
    new_ms = miller.set(crystal_symmetry=cs, indices=new_indices, anomalous_flag=True)
    data = flex.double(float(random.randrange(1, 100)) for _ in range(new_ms.size()))
    iobs = miller.array(new_ms, data, sigmas=data)
    iobs.change_symmetry(space_group_symbol="P222", merge_non_unique=False)
    iobs.set_info(miller.array_info(source="DIALS", source_type="reflection_tables"))
    iobs.set_observation_type_xray_intensity()

    plotter = AnomalousPlotter(iobs)
    d = plotter.make_plots()
    expected = ["normal_distribution_plot_highres", "anom_correl_plot"]
    keys = d.keys()
    for k in expected:
        assert k in keys
        assert d[k]["data"][0]["x"]  # check some data there

    plotter = AnomalousPlotter(iobs, strong_cutoff=3.0)
    d = plotter.make_plots()
    expected = [
        "anom_correl_plot",
        "anom_scatter_plot_lowres",
        "normal_distribution_plot_lowres",
    ]
    keys = list(d.keys())
    for k in expected:
        assert k in keys
        assert d[k]["data"][0]["x"]  # check some data there


def test_IntensityStatisticsPlots(iobs):
    n_bins = 2
    # mock the wilson scaling
    wilson_scaling = mock.Mock()
    wilson_scaling.d_star_sq = [1.0, 2.0]
    wilson_scaling.mean_I_obs_data = [1.0, 2.0]
    wilson_scaling.mean_I_obs_theory = [1.0, 2.0]
    wilson_scaling.mean_I_normalisation = [1.0, 2.0]
    wilson_scaling.iso_p_scale = 1.0
    # mock the twin results
    twin_results = mock.Mock()
    twin_results.nz_test.z = [1.0, 2.0]
    twin_results.nz_test.ac_obs = [1.0, 2.0]
    twin_results.nz_test.c_obs = [1.0, 2.0]
    twin_results.nz_test.ac_untwinned = [1.0, 2.0]
    twin_results.nz_test.c_untwinned = [1.0, 2.0]
    twin_results.l_test.l_values = [1.0, 2.0]
    twin_results.l_test.l_cumul_untwinned = [1.0, 2.0]
    twin_results.l_test.l_cumul_perfect_twin = [1.0, 2.0]
    twin_results.l_test.l_cumul = [1.0, 2.0]
    # mock the xtraige analysis
    xtriage_analyses = mock.Mock()
    xtriage_analyses.wilson_scaling = wilson_scaling
    xtriage_analyses.twin_results = twin_results
    xtriage_analyses.iso_b_wilson = 2.0

    plotter = IntensityStatisticsPlots(
        iobs, n_resolution_bins=n_bins, xtriage_analyses=xtriage_analyses
    )

    d = plotter.generate_resolution_dependent_plots()
    assert "wilson_intensity_plot" in d
    assert "second_moments" in d
    d = plotter.generate_miscellanous_plots()
    assert "cumulative_intensity_distribution" in d
    assert "l_test" in d

    def mock_xtriage(*args, **kwargs):
        return xtriage_analyses

    with mock.patch("mmtbx.scaling.xtriage.xtriage_analyses", new=mock_xtriage):
        plotter = IntensityStatisticsPlots(iobs)
        d = plotter.generate_resolution_dependent_plots()
        assert "wilson_intensity_plot" in d
        assert "second_moments" in d
        d = plotter.generate_miscellanous_plots()
        assert "cumulative_intensity_distribution" in d
        assert "l_test" in d

    # try with anomalous
    plotter = IntensityStatisticsPlots(
        iobs,
        anomalous=True,
        n_resolution_bins=n_bins,
        xtriage_analyses=xtriage_analyses,
    )

    d = plotter.generate_resolution_dependent_plots()
    assert "wilson_intensity_plot" in d
    assert "second_moments" in d
    d = plotter.generate_miscellanous_plots()
    assert "cumulative_intensity_distribution" in d
    assert "l_test" in d
    assert "multiplicities" in d


def test_i_over_sig_i_vs_i_plot(iobs):
    """Test the generation of 2d heatmap plots for intensity dist."""
    d = i_over_sig_i_vs_i_plot(iobs.data(), iobs.sigmas())
    assert "i_over_sig_i_vs_i" in d


def test_ResolutionPlotsAndStats(iobs):
    i_obs_anom = iobs.as_anomalous_array()
    iobs_anom = i_obs_anom.map_to_asu().customized_copy(info=iobs.info())
    n_bins = 2
    result = dataset_statistics(
        iobs, assert_is_not_unique_set_under_symmetry=False, n_bins=n_bins
    )
    anom_result = dataset_statistics(
        iobs_anom,
        assert_is_not_unique_set_under_symmetry=False,
        anomalous=True,
        n_bins=n_bins,
    )
    plotter = ResolutionPlotsAndStats(result, anom_result)

    assert plotter.d_star_sq_ticktext == ["1.74", "1.53", "1.38", "1.27", "1.18"]

    assert plotter.d_star_sq_tickvals == pytest.approx(
        [
            0.32984033277048164,
            0.42706274943714834,
            0.524285166103815,
            0.6215075827704818,
            0.7187299994371485,
        ],
        1e-4,
    )

    tables = plotter.statistics_tables()
    for expected in [
        "overall",
        "resolution_binned",
        "cc_half_significance_level",
        "overall_summary_data",
    ]:
        assert expected in tables  # overall and per resolution

    # test plots individually
    d = plotter.cc_one_half_plot()
    assert len(d["cc_one_half"]["data"]) == 6
    assert all(len(x["x"]) == n_bins for x in d["cc_one_half"]["data"][:4])
    d["cc_one_half"]["data"][0]["x"] == [
        0.5 * (uctbx.d_as_d_star_sq(b.d_max) + uctbx.d_as_d_star_sq(b.d_min))
        for b in result.bins
    ]

    d = plotter.i_over_sig_i_plot()
    assert len(d["i_over_sig_i"]["data"]) == 1
    assert len(d["i_over_sig_i"]["data"][0]["y"]) == n_bins

    d = plotter.r_pim_plot()
    assert len(d["r_pim"]["data"]) == 1
    assert len(d["r_pim"]["data"][0]["y"]) == n_bins

    d = plotter.completeness_plot()
    assert len(d["completeness"]["data"]) == 2
    assert len(d["completeness"]["data"][0]["y"]) == n_bins

    d = plotter.multiplicity_vs_resolution_plot()
    assert len(d["multiplicity_vs_resolution"]["data"]) == 2
    assert len(d["multiplicity_vs_resolution"]["data"][0]["y"]) == n_bins

    # now try centric options and sigma tau for cc_one_half
    plotter = ResolutionPlotsAndStats(result, anom_result, is_centric=True)
    d = plotter.cc_one_half_plot(method="sigma_tau")
    assert len(d["cc_one_half"]["data"]) == 6
    assert all(len(x["x"]) == n_bins for x in d["cc_one_half"]["data"][:2])
    assert d["cc_one_half"]["data"][2] == {}  # no anomalous plots
    assert d["cc_one_half"]["data"][3] == {}  # no anomalous plots
    assert d["cc_one_half"]["data"][4] == {}  # no cc_fit
    assert d["cc_one_half"]["data"][5] == {}  # no d_min
    d = plotter.completeness_plot()
    assert len(d["completeness"]["data"]) == 2
    assert len(d["completeness"]["data"][0]["y"]) == n_bins
    assert d["completeness"]["data"][1] == {}
    d = plotter.multiplicity_vs_resolution_plot()
    assert len(d["multiplicity_vs_resolution"]["data"]) == 2
    assert len(d["multiplicity_vs_resolution"]["data"][0]["y"]) == n_bins
    assert d["multiplicity_vs_resolution"]["data"][1] == {}

    plots = plotter.make_all_plots()
    assert list(plots.keys()) == [
        "cc_one_half",
        "i_over_sig_i",
        "completeness",
        "multiplicity_vs_resolution",
        "r_pim",
    ]
    for plot in plots.values():
        assert plot["layout"]["xaxis"]["ticktext"] == plotter.d_star_sq_ticktext
        assert plot["layout"]["xaxis"]["tickvals"] == plotter.d_star_sq_tickvals


@pytest.fixture
def batch_manager_fix():
    """Make a batch manager fixture"""

    batch_params = [{"id": 0, "range": [0, 10]}, {"id": 1, "range": [100, 110]}]
    batches = flex.int(itertools.chain(range(0, 10), range(100, 110)))
    return batch_manager(batches, batch_params)


def test_i_over_sig_i_vs_batch_plot(batch_manager_fix):
    """Test the IsigI batch plot"""
    bm = batch_manager_fix
    isigi = flex.double(range(0, 20))
    d = i_over_sig_i_vs_batch_plot(bm, isigi)
    assert list(d["i_over_sig_i_vs_batch"]["data"][0]["x"]) == list(bm.reduced_batches)
    assert list(d["i_over_sig_i_vs_batch"]["data"][0]["y"]) == list(isigi)


def test_scale_rmerge_vs_batch_plot(batch_manager_fix):
    """Test the scale and rmerge batch plot. Should have the option
    to plot without scales (for xia2)."""
    bm = batch_manager_fix
    rmergevsb = flex.double(range(0, 20))
    scalesvsb = flex.double(range(1, 21))
    d = scale_rmerge_vs_batch_plot(bm, rmergevsb, scalesvsb)
    assert list(d["scale_rmerge_vs_batch"]["data"][0]["x"]) == list(bm.reduced_batches)
    assert list(d["scale_rmerge_vs_batch"]["data"][1]["x"]) == list(bm.reduced_batches)
    assert list(d["scale_rmerge_vs_batch"]["data"][0]["y"]) == list(scalesvsb)
    assert list(d["scale_rmerge_vs_batch"]["data"][1]["y"]) == list(rmergevsb)

    d = scale_rmerge_vs_batch_plot(bm, rmergevsb)
    assert d["scale_rmerge_vs_batch"]["data"][0] == {}
    assert list(d["scale_rmerge_vs_batch"]["data"][1]["x"]) == list(bm.reduced_batches)
    assert d["scale_rmerge_vs_batch"]["data"][0] == {}
    assert list(d["scale_rmerge_vs_batch"]["data"][1]["y"]) == list(rmergevsb)



# LIBTBX_PRE_DISPATCHER_INCLUDE_SH export PHENIX_GUI_ENVIRONMENT=1

from __future__ import annotations

from dials.algorithms.shoebox import MaskCode
from dials.array_family import flex
from dials.viewer.slice_viewer import show_3d

if __name__ == "__main__":
    lst_flex = []
    lst_flex_norm = []

    for size_xyz in range(8, 6, -1):

        size_x = size_xyz * 2

        data_xyz_flex = flex.double(flex.grid(size_xyz, size_xyz, size_x), 15)
        data_flex_norm = flex.double(flex.grid(size_xyz, size_xyz, size_x), 15)
        data_flex_mask = flex.int(flex.grid(size_xyz, size_xyz, size_x), 0)

        tot = 0.0
        for frm in range(size_xyz):
            for row in range(size_xyz):
                for col in range(size_x):
                    data_xyz_flex[frm, row, col] += row * 2 + col * 2 + frm * 2
                    tot += data_xyz_flex[frm, row, col]
                    if row > 1 and row < size_xyz - 2 and col > 1 and col < size_x - 2:
                        data_flex_mask[frm, row, col] = MaskCode.Foreground

                        different_mask_values = """
            MaskCode.Valid           =  "\\\\\\"
            MaskCode.Foreground      =  "//////"
            MaskCode.Background      =  "||||||"
            MaskCode.BackgroundUsed  =  "------"
            """

        for frm in range(size_xyz):
            for row in range(size_xyz):
                for col in range(size_x):
                    data_flex_norm[frm, row, col] += data_xyz_flex[frm, row, col] / tot

        lst_flex.append(data_xyz_flex)
        lst_flex_norm.append(data_flex_norm)

    show_3d(data_xyz_flex, data_flex_mask)
    show_3d(data_xyz_flex)
    show_3d(lst_flex)
    show_3d(lst_flex_norm)
    print("Test Done")



from __future__ import annotations

import random

from dials.algorithms.shoebox import find_overlapping


def test_single_panel():
    from dials.array_family import flex

    nrefl = 1000

    # Generate bboxes
    bbox = flex.int6(nrefl)
    for i in range(nrefl):
        x0 = random.randint(0, 500)
        y0 = random.randint(0, 500)
        z0 = random.randint(0, 10)
        x1 = x0 + random.randint(2, 10)
        y1 = y0 + random.randint(2, 10)
        z1 = z0 + random.randint(2, 10)
        bbox[i] = (x0, x1, y0, y1, z0, z1)

    # Find the overlaps
    overlaps = find_overlapping(bbox)

    assert overlaps.num_vertices() == nrefl
    overlaps2 = brute_force(bbox)
    assert overlaps.num_edges() == len(overlaps2)
    edges = {}
    for edge in overlaps2:
        edge = (min(edge), max(edge))
        edges[edge] = None
    for edge in overlaps.edges():
        edge = overlaps.source(edge), overlaps.target(edge)
        edge = (min(edge), max(edge))
        assert edge in edges


def test_multiple_panels():
    from dials.array_family import flex

    nrefl = 1000

    # Generate bboxes
    bbox = flex.int6(nrefl)
    panel = flex.size_t(nrefl)
    for i in range(nrefl):
        x0 = random.randint(0, 500)
        y0 = random.randint(0, 500)
        z0 = random.randint(0, 10)
        x1 = x0 + random.randint(2, 10)
        y1 = y0 + random.randint(2, 10)
        z1 = z0 + random.randint(2, 10)
        bbox[i] = (x0, x1, y0, y1, z0, z1)
        panel[i] = random.randint(0, 2)

    # Find the overlaps
    overlaps = find_overlapping(bbox, panel)
    assert overlaps.num_vertices() == nrefl
    overlaps2 = brute_force(bbox, panel)
    assert overlaps.num_edges() == len(overlaps2)
    edges = {}
    for edge in overlaps2:
        edge = (min(edge), max(edge))
        edges[edge] = None
    for edge in overlaps.edges():
        edge = (overlaps.source(edge), overlaps.target(edge))
        edge = (min(edge), max(edge))
        assert edge in edges


def brute_force(bbox, panel=None):
    overlaps = []
    if panel is None:
        for j in range(len(bbox) - 1):
            jx0, jx1, jy0, jy1, jz0, jz1 = bbox[j]
            for i in range(j + 1, len(bbox)):
                ix0, ix1, iy0, iy1, iz0, iz1 = bbox[i]
                if not (
                    ix0 >= jx1
                    or jx0 >= ix1
                    or iy0 >= jy1
                    or jy0 >= iy1
                    or iz0 >= jz1
                    or jz0 >= iz1
                ):
                    overlaps.append((i, j))
    else:
        for j in range(len(bbox) - 1):
            jx0, jx1, jy0, jy1, jz0, jz1 = bbox[j]
            for i in range(j + 1, len(bbox)):
                ix0, ix1, iy0, iy1, iz0, iz1 = bbox[i]
                if panel[j] != panel[i]:
                    continue
                if not (
                    ix0 >= jx1
                    or jx0 >= ix1
                    or iy0 >= jy1
                    or jy0 >= iy1
                    or iz0 >= jz1
                    or jz0 >= iz1
                ):
                    overlaps.append((i, j))

    return overlaps



from __future__ import annotations

from dials.array_family import flex
from dials.viewer.slice_viewer import show_reflections


def extract_n_show(table):
    show_reflections(table, two_windows=True)


if __name__ == "__main__":
    import sys

    pick_name = sys.argv[1]

    table = flex.reflection_table.from_file(pick_name)
    extract_n_show(table)



from __future__ import annotations

import concurrent.futures
import copy
import logging
import math

import libtbx
import libtbx.phil
import scitbx.matrix
from cctbx import crystal, miller, sgtbx
from cctbx.crystal_orientation import crystal_orientation
from cctbx.sgtbx import bravais_types
from dxtbx.model import Crystal
from libtbx.introspection import number_of_processors
from rstbx.dps_core.lepage import iotbx_converter
from rstbx.symmetry.subgroup import MetricSubgroup
from scitbx.array_family import flex

import dials.util
from dials.algorithms.indexing.refinement import refine
from dials.command_line.check_indexing_symmetry import (
    get_symop_correlation_coefficients,
)
from dials.util.log import LoggingContext

logger = logging.getLogger(__name__)


phil_scope = libtbx.phil.parse(
    """
lepage_max_delta = 5
  .type = float
nproc = Auto
  .type = int(value_min=1)
cc_n_bins = None
  .type = int(value_min=1)
  .help = "Number of resolution bins to use for calculation of correlation coefficients"

best_monoclinic_beta = True
  .type = bool
  .help = "If True, then for monoclinic centered cells, I2 will be preferred over C2 if"
          "it gives a less oblique cell (i.e. smaller beta angle)."

include scope dials.algorithms.refinement.refiner.phil_scope
""",
    process_includes=True,
)

# override default refinement parameters
phil_scope = phil_scope.fetch(
    source=libtbx.phil.parse(
        """\
refinement {
  reflections {
    reflections_per_degree=100
  }
}
"""
    )
)


def dxtbx_crystal_from_orientation(crystal_orientation, space_group):
    """Convert a cctbx crystal_orientation to a dxtbx.Crystal model.

    Args:
        crystal_orientation (cctbx.crystal_orientation):
            A cctbx crystal_orientation object
        space_group (cctbx.sgtbx.space_group): The space group.

    Returns:
        dxtbx.model.Crystal: The dxtbx crystal model.
    """
    dm = crystal_orientation.direct_matrix()
    AA = scitbx.matrix.col((dm[0], dm[1], dm[2]))
    BB = scitbx.matrix.col((dm[3], dm[4], dm[5]))
    CC = scitbx.matrix.col((dm[6], dm[7], dm[8]))

    cryst = Crystal(
        real_space_a=AA, real_space_b=BB, real_space_c=CC, space_group=space_group
    )
    return cryst


class BravaisSetting(MetricSubgroup):  # inherits from dictionary
    def __init__(self, other):
        self.update(other)


class RefinedSettingsList(list):
    def __init__(self, *args, cb_op_to_primitive=None, **kwargs):
        if cb_op_to_primitive is None:
            cb_op_to_primitive = sgtbx.change_of_basis_op()
        self.cb_op_to_primitive = cb_op_to_primitive
        super().__init__(*args, **kwargs)

    def supergroup(self):
        return self[0]

    def triclinic(self):
        return self[-1]

    def as_dict(self):
        result = {}

        for item in self:
            uc = item.refined_crystal.get_unit_cell()
            result[item.setting_number] = {
                "max_angular_difference": item["max_angular_difference"],
                "rmsd": item.rmsd,
                "nspots": item.Nmatches,
                "bravais": item["bravais"],
                "unit_cell": uc.parameters(),
                "cb_op": str(item["cb_op_inp_best"] * self.cb_op_to_primitive),
                "max_cc": item.max_cc,
                "min_cc": item.min_cc,
                "correlation_coefficients": list(item.correlation_coefficients),
                "cc_nrefs": list(item.cc_nrefs),
                "recommended": item.recommended,
            }

        return result

    def __str__(self):
        table_data = [
            [
                "Solution",
                "Metric fit",
                "rmsd",
                "min/max cc",
                "#spots",
                "lattice",
                "unit_cell",
                "volume",
                "cb_op",
            ]
        ]
        for item in self:
            uc = item.refined_crystal.get_unit_cell()
            P = uc.parameters()
            min_max_cc_str = "-/-"
            if item.min_cc is not None and item.max_cc is not None:
                min_max_cc_str = f"{item.min_cc:.3f}/{item.max_cc:.3f}"
            if item.recommended:
                status = "*"
            else:
                status = ""
            table_data.append(
                [
                    "%1s%7d" % (status, item.setting_number),
                    f"{item['max_angular_difference']:6.4f}",
                    f"{item.rmsd:5.3f}",
                    min_max_cc_str,
                    "%d" % item.Nmatches,
                    f"{item['bravais']}",
                    "%6.2f %6.2f %6.2f %6.2f %6.2f %6.2f" % P,
                    f"{uc.volume():.0f}",
                    f"{str(item['cb_op_inp_best'] * self.cb_op_to_primitive)}",
                ]
            )

        output = dials.util.tabulate(
            table_data, headers="firstrow", colalign=("right",)
        )
        output = output + "\n* = recommended solution\n"
        return output


# Mapping of Bravais lattice type to corresponding lowest possible symmetry
bravais_lattice_to_lowest_symmetry_spacegroup_number = {
    "aP": 1,
    "mP": 3,
    "mC": 5,
    "oP": 16,
    "oC": 20,
    "oF": 22,
    "oI": 23,
    "tP": 75,
    "tI": 79,
    "hP": 143,
    "hR": 146,
    "cP": 195,
    "cF": 196,
    "cI": 197,
}


def lowest_symmetry_space_group_for_bravais_lattice(
    bravais_lattice: str,
) -> sgtbx.space_group:
    if bravais_lattice == "mI":
        return sgtbx.space_group_info("I2").group()
    return sgtbx.space_group_info(
        number=bravais_lattice_to_lowest_symmetry_spacegroup_number[bravais_lattice]
    ).group()


def refined_settings_from_refined_triclinic(
    experiments, reflections, params, cb_op_to_primitive=None
):
    """Generate a RefinedSettingsList from a triclinic model.

    Args:
        experiments: The experiments refined with a triclinic model
        reflections: A reflection table containing observed centroids
        params: The working PHIL parameters.

    Returns:
        RefinedSettingsList: A list of the refined settings. The highest symmetry
        setting will be first item in the list, and the triclinic setting will be last.
    """

    if params.nproc is libtbx.Auto:
        params.nproc = number_of_processors()

    if params.refinement.reflections.outlier.algorithm in ("auto", libtbx.Auto):
        if experiments[0].goniometer is None:
            params.refinement.reflections.outlier.algorithm = "sauter_poon"
        else:
            # different default to dials.refine
            # tukey is faster and more appropriate at the indexing step
            params.refinement.reflections.outlier.algorithm = "tukey"

    assert len(experiments.crystals()) == 1
    crystal = experiments.crystals()[0]

    used_reflections = copy.deepcopy(reflections)
    UC = crystal.get_unit_cell()

    refined_settings = RefinedSettingsList(cb_op_to_primitive=cb_op_to_primitive)
    for item in iotbx_converter(
        UC, params.lepage_max_delta, best_monoclinic_beta=params.best_monoclinic_beta
    ):
        refined_settings.append(BravaisSetting(item))

    triclinic = refined_settings.triclinic()

    # assert no transformation between indexing and bravais list
    assert str(triclinic["cb_op_inp_best"]) == "a,b,c"

    Nset = len(refined_settings)
    for j in range(Nset):
        refined_settings[j].setting_number = Nset - j

    for subgroup in refined_settings:
        bravais_lattice = str(
            bravais_types.bravais_lattice(group=subgroup["best_subsym"].space_group())
        )
        space_group = lowest_symmetry_space_group_for_bravais_lattice(bravais_lattice)
        orient = crystal_orientation(crystal.get_A(), True).change_basis(
            scitbx.matrix.sqr(
                subgroup["cb_op_inp_best"].c().as_double_array()[0:9]
            ).transpose()
        )
        constrain_orient = orient.constrain(subgroup["system"])
        subgroup["bravais"] = bravais_lattice
        subgroup.unrefined_crystal = dxtbx_crystal_from_orientation(
            constrain_orient, space_group
        )

    with concurrent.futures.ProcessPoolExecutor(max_workers=params.nproc) as pool:
        for i, result in enumerate(
            pool.map(
                refine_subgroup,
                (
                    (params, subgroup, used_reflections, experiments)
                    for subgroup in refined_settings
                ),
            )
        ):
            refined_settings[i] = result

    identify_likely_solutions(refined_settings)
    return refined_settings


def identify_likely_solutions(all_solutions):
    """Identify likely solutions using heuristics.

    Args:
        all_solutions (RefinedSettingsList): The list of refined bravais settings.

    Use a set of heuristics to identify likely solutions, by comparing refined rmsds
    in a given setting with the triclinic rmsds. Also looks at the max_angular
    difference and the correlation coefficients for the solutions. Sets the
    `recommended` attribute of each solution to `True` or `False` as appropriate.
    """
    p1_solution = all_solutions[-1]
    assert p1_solution.setting_number == 1, p1_solution.setting_number
    rmsd_p1 = p1_solution.rmsd

    for solution in all_solutions:
        solution.recommended = False
        if solution["max_angular_difference"] < 0.5:
            if (
                solution.min_cc is None or solution.min_cc < 0.5
            ) and solution.rmsd > 1.5 * rmsd_p1:
                continue
        elif (
            solution.min_cc is None or solution.min_cc < 0.7
        ) and solution.rmsd > 2.0 * rmsd_p1:
            continue
        elif solution.rmsd > 3 * rmsd_p1:
            continue
        solution.recommended = True


def refine_subgroup(args):
    assert len(args) == 4
    params, subgroup, used_reflections, experiments = args

    used_reflections = copy.deepcopy(used_reflections)
    triclinic_miller = used_reflections["miller_index"]
    higher_symmetry_miller = subgroup["cb_op_inp_best"].apply(triclinic_miller)
    used_reflections["miller_index"] = higher_symmetry_miller
    unrefined_crystal = copy.deepcopy(subgroup.unrefined_crystal)
    for expt in experiments:
        expt.crystal = unrefined_crystal

    subgroup.max_cc = None
    subgroup.min_cc = None
    subgroup.correlation_coefficients = []
    subgroup.cc_nrefs = []

    with LoggingContext("dials.algorithms.refinement", level=logging.ERROR):
        try:
            outlier_algorithm = params.refinement.reflections.outlier.algorithm
            sel = used_reflections.get_flags(used_reflections.flags.used_in_refinement)
            if sel.all_eq(False):
                # Soft outlier rejection if no used_in_refinement flag is set
                params.refinement.reflections.outlier.algorithm = "tukey"
                iqr_multiplier = (
                    params.refinement.reflections.outlier.tukey.iqr_multiplier
                )
                params.refinement.reflections.outlier.tukey.iqr_multiplier = (
                    2 * iqr_multiplier
                )
                sel = ~sel
            else:
                # Remove reflections not previously used in refinement
                params.refinement.reflections.outlier.algorithm = "null"
            refinery, refined, outliers = refine(
                params, used_reflections.select(sel), experiments
            )
            params.refinement.reflections.outlier.algorithm = outlier_algorithm
            refinery, refined, outliers = refine(
                params, used_reflections, refinery.get_experiments()
            )
        except RuntimeError as e:
            if (
                str(e) == "scitbx Error: g0 - astry*astry -astrz*astrz <= 0."
                or str(e) == "scitbx Error: g1-bstrz*bstrz <= 0."
            ):
                subgroup.refined_experiments = None
                subgroup.rmsd = None
                subgroup.Nmatches = None
                return subgroup
            else:
                raise
        dall = refinery.rmsds()
        dx = dall[0]
        dy = dall[1]
        subgroup.rmsd = math.sqrt(dx * dx + dy * dy)
        subgroup.Nmatches = len(refinery.get_matches())
        subgroup.refined_experiments = refinery.get_experiments()
        assert len(subgroup.refined_experiments.crystals()) == 1
        subgroup.refined_crystal = subgroup.refined_experiments.crystals()[0]
        cs = crystal.symmetry(
            unit_cell=subgroup.refined_crystal.get_unit_cell(),
            space_group=subgroup.refined_crystal.get_space_group(),
        )
        if "intensity.sum.value" in used_reflections:
            # remove refl with -ve variance
            sel = used_reflections["intensity.sum.variance"] > 0
            good_reflections = used_reflections.select(sel)

            ms = miller.set(cs, good_reflections["miller_index"])
            ms = ms.array(
                good_reflections["intensity.sum.value"]
                / flex.sqrt(good_reflections["intensity.sum.variance"])
            )
            if params.cc_n_bins is not None:
                ms.setup_binner(n_bins=params.cc_n_bins)
            ccs, nrefs = get_symop_correlation_coefficients(
                ms, use_binning=(params.cc_n_bins is not None)
            )
            subgroup.correlation_coefficients = ccs
            subgroup.cc_nrefs = nrefs
            ccs = ccs.select(nrefs > 10)
            if len(ccs) > 1:
                subgroup.max_cc = flex.max(ccs[1:])
                subgroup.min_cc = flex.min(ccs[1:])
    return subgroup



from __future__ import annotations

import procrunner
import pytest

from dials.array_family import flex


def test_against_dials_integrate(dials_data, tmp_path):

    ## ensure insulin folder exists
    dials_data("insulin", pathlib=True)

    # Run as a single job to avoid splitting reflections
    procrunner.run(
        (
            "dials.integrate",
            dials_data("insulin_processed", pathlib=True) / "refined.expt",
            dials_data("insulin_processed", pathlib=True) / "refined.refl",
            "mp.njobs=1",
            "mp.nproc=1",
        ),
        working_directory=tmp_path,
    ).check_returncode()

    procrunner.run(
        (
            "dev.dials.simple_integrate",
            dials_data("insulin_processed", pathlib=True) / "refined.expt",
            dials_data("insulin_processed", pathlib=True) / "refined.refl",
        ),
        working_directory=tmp_path,
    ).check_returncode()

    simple_refl = flex.reflection_table.from_msgpack_file(
        tmp_path / "simple_integrated.refl"
    )
    dials_refl = flex.reflection_table.from_msgpack_file(tmp_path / "integrated.refl")

    matches = dials_refl.match_with_reference(simple_refl)[0]
    dials_refl = dials_refl.select(matches)

    assert len(simple_refl) == len(dials_refl)
    assert sum(
        simple_refl["intensity.sum.value"] - dials_refl["intensity.sum.value"]
    ) == pytest.approx(0.0, abs=1e-7)
    assert sum(
        simple_refl["intensity.prf.value"] - dials_refl["intensity.prf.value"]
    ) == pytest.approx(0.0, abs=1e-7)
    assert sum(
        simple_refl["background.sum.value"] - dials_refl["background.sum.value"]
    ) == pytest.approx(0.0, abs=1e-7)
    assert sum(
        simple_refl["background.mean"] - dials_refl["background.mean"]
    ) == pytest.approx(0.0, abs=1e-7)
    assert sum(
        simple_refl["background.dispersion"] - dials_refl["background.dispersion"]
    ) == pytest.approx(0.0, abs=1e-7)
    assert sum(
        simple_refl["background.mse"] - dials_refl["background.mse"]
    ) == pytest.approx(0.0, abs=1e-7)
    assert sum(
        simple_refl["background.sum.variance"] - dials_refl["background.sum.variance"]
    ) == pytest.approx(0.0, abs=1e-7)
    assert sum(
        simple_refl["intensity.prf.variance"] - dials_refl["intensity.prf.variance"]
    ) == pytest.approx(0.0, abs=1e-7)
    assert sum(
        simple_refl["intensity.sum.variance"] - dials_refl["intensity.sum.variance"]
    ) == pytest.approx(0.0, abs=1e-7)
    assert sum(simple_refl["lp"] - dials_refl["lp"]) == pytest.approx(0.0, abs=1e-7)
    assert sum(simple_refl["zeta"] - dials_refl["zeta"]) == pytest.approx(0.0, abs=1e-7)
    assert sum(simple_refl["d"] - dials_refl["d"]) == pytest.approx(0.0, abs=1e-7)
    assert sum(simple_refl["partiality"] - dials_refl["partiality"]) == pytest.approx(
        0.0, abs=1e-7
    )
    assert (
        simple_refl["num_pixels.foreground"] == dials_refl["num_pixels.foreground"]
    ).count(True) == len(simple_refl)
    assert (
        simple_refl["num_pixels.background"] == dials_refl["num_pixels.background"]
    ).count(True) == len(simple_refl)
    assert (simple_refl["num_pixels.valid"] == dials_refl["num_pixels.valid"]).count(
        True
    ) == len(simple_refl)
    assert (
        simple_refl["num_pixels.background_used"]
        == dials_refl["num_pixels.background_used"]
    ).count(True) == len(simple_refl)
    assert sum(
        simple_refl["profile.correlation"] - dials_refl["profile.correlation"]
    ) == pytest.approx(0.0, abs=1e-7)



# LIBTBX_PRE_DISPATCHER_INCLUDE_SH export PHENIX_GUI_ENVIRONMENT=1

from __future__ import annotations

import json
import math
import os
import sys

import matplotlib

import iotbx.phil
from cctbx import crystal, miller
from cctbx.array_family import flex
from scitbx import matrix

import dials.util

help_message = """

Calculates a stereographic projection image for the given crystal models and
the given miller indices (either specified individually, or for all miller indices
up to a given hkl_limit). By default the projection is in the plane
perpendicular to 0,0,1 reflection for the first crystal, however the projection
can optionally be performed in the laboratory frame (frame=laboratory) in the
plane perpendicular to the beam. Setting the parameter expand_to_p1=True will
also plot all symmetry equivalents of the given miller indices, and
eliminate_sys_absent=False will eliminate systematically absent reflections
before generating the projection.

Examples::

  dials.stereographic_projection indexed.expt hkl=1,0,0 hkl=0,1,0

  dials.stereographic_projection indexed.expt hkl_limit=2

  dials.stereographic_projection indexed_1.expt indexed_2.expt hkl=1,0,0 expand_to_p1=True
"""

phil_scope = iotbx.phil.parse(
    """
hkl = None
  .type = ints(size=3)
  .multiple=True
hkl_limit = None
  .type = int(value_min=1)
expand_to_p1 = True
  .type = bool
  .help = "Expand the given miller indices to symmetry equivalent reflections"
eliminate_sys_absent = False
  .type = bool
  .help = "Eliminate systematically absent reflections"
frame = *laboratory crystal
  .type = choice
phi_angle = 0
  .type = float
  .help = "Phi rotation angle (degrees)"
use_starting_angle = False
  .type = bool
  .help = "If True, then the projection will be done for each crystal at the "
          "starting phi angle for the scan associated with the crystal."
plane_normal = None
  .type = ints(size=3)
save_coordinates = True
  .type = bool
plot {
  filename = stereographic_projection.png
    .type = path
  label_indices = False
    .type = bool
  colours = None
    .type = strings
  marker_size = 3
    .type = int(value_min=1)
  font_size = 6
    .type = float(value_min=0)
  colour_map = None
    .type = str
  gridsize = None
    .type = int
  labels = None
    .type = strings
}
json {
  filename = None
    .type = path
}
"""
)


def reference_poles_perpendicular_to_beam(beam, goniometer):
    # plane normal
    d0 = matrix.col(beam.get_s0()).normalize()
    if goniometer is not None:
        d1 = d0.cross(matrix.col(goniometer.get_rotation_axis())).normalize()
    else:
        d1 = d0.ortho()
    d2 = d1.cross(d0).normalize()
    return (d0, d1, d2)


def reference_poles_crystal(crystal_model, plane_normal=(0, 0, 1)):
    A = matrix.sqr(crystal_model.get_A())
    B = matrix.sqr(crystal_model.get_B())
    A_inv = A.inverse()
    G = A_inv * A_inv.transpose()
    G_star = A.transpose() * A
    h0 = (G * matrix.col(plane_normal)).normalize()
    h1 = matrix.col((1, 0, 0)).cross((G_star * h0).normalize())
    h2 = (G_star * h1).cross(G_star * h0).normalize()
    return tuple((B * h).normalize() for h in (h0, h1, h2))


def stereographic_projection(points, reference_poles):
    # https://doi.org/10.1107/S0021889868005029
    # J. Appl. Cryst. (1968). 1, 68-70
    # The construction of stereographic projections by computer
    # G. K. Stokes, S. R. Keown and D. J. Dyson

    assert len(reference_poles) == 3
    r_0, r_1, r_2 = reference_poles

    projections = flex.vec2_double()

    for p in points:
        r_i = matrix.col(p)
        # theta is the angle between r_i and the plane normal, r_0
        cos_theta = r_i.cos_angle(r_0)
        if cos_theta < 0:
            r_i = -r_i
            cos_theta = r_i.cos_angle(r_0)

        # alpha is the angle between r_i and r_1
        cos_alpha = r_i.cos_angle(r_1)
        theta = math.acos(cos_theta)
        cos_phi = cos_alpha / math.sin(theta)
        if abs(cos_phi) > 1:
            cos_phi = math.copysign(1, cos_phi)
        phi = math.acos(cos_phi)

        N = r_i.dot(r_2)
        r = math.tan(theta / 2)
        x = r * cos_phi
        y = r * math.sin(phi)
        y = math.copysign(y, N)

        projections.append((x, y))

    return projections


def gcd_list(l):
    # greatest common divisor for a list of numbers
    from scitbx.math import gcd_int_simple as gcd

    result = l[0]
    for i in range(1, len(l)):
        result = gcd(result, l[i])
    return result


@dials.util.show_mail_handle_errors()
def run(args=None):
    from dials.util.options import ArgumentParser, flatten_experiments

    # The script usage
    usage = "dials.stereographic_projection [options] [param.phil] indexed.expt"

    parser = ArgumentParser(
        usage=usage,
        phil=phil_scope,
        read_experiments=True,
        check_format=False,
        epilog=help_message,
    )

    params, options = parser.parse_args(args=args, show_diff_phil=True)
    experiments = flatten_experiments(params.input.experiments)

    if not experiments:
        parser.print_help()
        return

    if not params.hkl and params.hkl_limit is None:
        sys.exit("Please provide hkl or hkl_limit parameters.")

    if params.plot.labels and len(params.plot.labels) != len(experiments):
        sys.exit(
            "Number of labels (%i) must equal number of experiments (%i)"
            % (len(params.plot.labels), len(experiments))
        )

    if params.hkl is not None and len(params.hkl):
        miller_indices = flex.miller_index(params.hkl)
    elif params.hkl_limit is not None:
        limit = params.hkl_limit
        miller_indices = flex.miller_index()
        for h in range(-limit, limit + 1):
            for k in range(-limit, limit + 1):
                for l in range(-limit, limit + 1):
                    if (h, k, l) == (0, 0, 0):
                        continue
                    miller_indices.append((h, k, l))

    crystals = experiments.crystals()

    symmetry = crystal.symmetry(
        unit_cell=crystals[0].get_unit_cell(), space_group=crystals[0].get_space_group()
    )
    miller_set = miller.set(symmetry, miller_indices)
    d_spacings = miller_set.d_spacings()
    if params.eliminate_sys_absent:
        d_spacings = d_spacings.eliminate_sys_absent()
    if params.expand_to_p1:
        d_spacings = d_spacings.as_non_anomalous_array().expand_to_p1()
    miller_indices = d_spacings.indices()

    # find the greatest common factor (divisor) between miller indices
    miller_indices_unique = flex.miller_index()
    for hkl in miller_indices:
        gcd = gcd_list(hkl)
        if gcd > 1:
            miller_indices_unique.append(tuple(int(h / gcd) for h in hkl))
        elif gcd < 1:
            pass
        else:
            miller_indices_unique.append(hkl)
    miller_indices = miller_indices_unique
    miller_indices = flex.miller_index(list(set(miller_indices)))

    ref_crystal = crystals[0]
    U = matrix.sqr(ref_crystal.get_U())
    B = matrix.sqr(ref_crystal.get_B())
    R = matrix.identity(3)

    if params.frame == "laboratory":
        reference_poles = reference_poles_perpendicular_to_beam(
            experiments[0].beam, experiments[0].goniometer
        )
        if params.use_starting_angle:
            rotation_axis = matrix.col(experiments[0].goniometer.get_rotation_axis())
            R = rotation_axis.axis_and_angle_as_r3_rotation_matrix(
                experiments[0].scan.get_oscillation()[0], deg=True
            )
        elif params.phi_angle != 0:
            rotation_axis = matrix.col(experiments[0].goniometer.get_rotation_axis())
            R = rotation_axis.axis_and_angle_as_r3_rotation_matrix(
                params.phi_angle, deg=True
            )
    else:
        if params.plane_normal is not None:
            plane_normal = params.plane_normal
        else:
            plane_normal = (0, 0, 1)
        reference_poles = reference_poles_crystal(
            ref_crystal, plane_normal=plane_normal
        )

    if params.frame == "crystal":
        U = matrix.identity(3)

    reciprocal_space_points = list(R * U * B) * miller_indices.as_vec3_double()
    projections_ref = stereographic_projection(reciprocal_space_points, reference_poles)

    projections_all = [projections_ref]

    if experiments:
        from dials.algorithms.indexing.compare_orientation_matrices import (
            difference_rotation_matrix_axis_angle,
        )

        for expt in experiments[1:]:
            cryst = expt.crystal
            if params.frame == "crystal":
                R_ij, axis, angle, cb_op = difference_rotation_matrix_axis_angle(
                    ref_crystal, cryst
                )
                U = R_ij
            elif params.use_starting_angle:
                if params.use_starting_angle:
                    rotation_axis = matrix.col(expt.goniometer.get_rotation_axis())
                    R = rotation_axis.axis_and_angle_as_r3_rotation_matrix(
                        expt.scan.get_oscillation()[0], deg=True
                    )
            else:
                U = matrix.sqr(cryst.get_U())
            reciprocal_space_points = (
                list(R * U * matrix.sqr(cryst.get_B()))
                * miller_indices.as_vec3_double()
            )
            projections = stereographic_projection(
                reciprocal_space_points, reference_poles
            )
            projections_all.append(projections)

    if params.save_coordinates:
        with open("projections.txt", "w") as f:
            f.write("crystal h k l x y" + os.linesep)
            for i_cryst, projections in enumerate(projections_all):
                for hkl, proj in zip(miller_indices, projections):
                    f.write("%i " % (i_cryst + 1))
                    f.write("%i %i %i " % hkl)
                    f.write(("%f %f" + os.linesep) % proj)

    if params.plot.filename:
        epochs = None
        if params.plot.colour_map is not None:
            if experiments[0].scan is not None:
                epochs = [expt.scan.get_epochs()[0] for expt in experiments]
            else:
                epochs = [i for i, expt in enumerate(experiments)]
        plot_projections(
            projections_all,
            filename=params.plot.filename,
            colours=params.plot.colours,
            marker_size=params.plot.marker_size,
            font_size=params.plot.font_size,
            gridsize=params.plot.gridsize,
            label_indices=miller_indices if params.plot.label_indices else False,
            epochs=epochs,
            colour_map=params.plot.colour_map,
        )

    if params.json.filename:
        projections_as_json(
            projections_all, filename=params.json.filename, labels=params.plot.labels
        )


def plot_projections(
    projections,
    filename=None,
    colours=None,
    marker_size=3,
    font_size=6,
    gridsize=None,
    label_indices=False,
    epochs=None,
    colour_map=None,
):
    projections_all = projections

    # http://matplotlib.org/faq/howto_faq.html#generate-images-without-having-a-window-appear
    matplotlib.use("Agg")  # use a non-interactive backend
    from matplotlib import pylab, pyplot

    if epochs is not None and colour_map is not None:
        epochs = flex.double(epochs)
        epochs -= flex.min(epochs)
        epochs /= flex.max(epochs)
        cmap = matplotlib.cm.get_cmap(colour_map)
        colours = [cmap(e) for e in epochs]
    elif colours is None or len(colours) == 0:
        colours = ["b"] * len(projections_all)
    elif len(colours) < len(projections_all):
        colours = colours * len(projections_all)

    fig = pyplot.figure()

    pyplot.scatter([0], [0], marker="+", c="0.75", s=100)
    cir = pylab.Circle((0, 0), radius=1.0, fill=False, color="0.75")
    pylab.gca().add_patch(cir)

    if gridsize is not None:
        x = flex.double()
        y = flex.double()
        for i, projections in enumerate(projections_all):
            x_, y_ = projections.parts()
            x.extend(x_)
            y.extend(y_)
        hb = pyplot.hexbin(x, y, gridsize=gridsize, linewidths=0.2)
        pyplot.colorbar(hb)
    else:
        for i, projections in enumerate(projections_all):
            x, y = projections.parts()
            pyplot.scatter(
                x.as_numpy_array(),
                y.as_numpy_array(),
                c=colours[i],
                s=marker_size,
                edgecolors="none",
            )
            if label_indices:
                for j, (hkl, proj) in enumerate(zip(label_indices, projections)):
                    # hack to not write two labels on top of each other
                    p1, p2 = (projections - proj).parts()
                    if (flex.sqrt(flex.pow2(p1) + flex.pow2(p2)) < 1e-3).iselection()[
                        0
                    ] != j:
                        continue
                    pyplot.text(proj[0], proj[1], str(hkl), fontsize=font_size)
    fig.axes[0].set_aspect("equal")
    pyplot.xlim(-1.1, 1.1)
    pyplot.ylim(-1.1, 1.1)
    if filename is not None:
        pyplot.savefig(filename, dpi=300)


def projections_as_dict(projections, labels):
    projections_all = flex.vec2_double()
    if labels:
        labels_all = []
        assert len(projections) == len(labels)
    for i, proj in enumerate(projections):
        projections_all.extend(proj)
        if labels:
            labels_all.extend([labels[i]] * len(proj))

    data = []
    x, y = projections_all.parts()
    data.append(
        {
            "x": list(x),
            "y": list(y),
            "mode": "markers",
            "type": "scatter",
            "name": "stereographic_projections",
            "showlegend": False,
            "hovertext": labels_all if labels else "",
            "hoverinfo": "text",
        }
    )
    data.append(
        {
            "x": [0],
            "y": [0],
            "mode": "markers",
            "marker": {
                "color": "black",
                "size": 25,
                "symbol": "cross-thin",
                "line": {"width": 1},
            },
            "showlegend": False,
        }
    )

    d = {
        "data": data,
        "layout": {
            "title": "Stereographic projections",
            "hovermode": "closest",
            "xaxis": {
                "range": [-1.0, 1.0],
                "showgrid": False,
                "zeroline": False,
                "showline": False,
                "ticks": "",
                "showticklabels": False,
            },
            "yaxis": {
                "range": [-1.0, 1.0],
                "showgrid": False,
                "zeroline": False,
                "showline": False,
                "ticks": "",
                "showticklabels": False,
            },
            "shapes": [
                {
                    "type": "circle",
                    "xref": "x",
                    "yref": "y",
                    "x0": -1,
                    "y0": -1,
                    "x1": 1,
                    "y1": 1,
                    "line": {"color": "black"},
                }
            ],
        },
    }
    return d


def projections_as_json(projections, filename=None, labels=None):
    d = projections_as_dict(projections, labels=labels)

    json_str = json.dumps(d)
    if filename is not None:
        with open(filename, "w") as f:
            f.write(json_str)
    return json_str


if __name__ == "__main__":
    run()



"""Setup experimental geometry for refinement test cases"""


from __future__ import annotations

from libtbx.phil import command_line, parse

from dials.algorithms.refinement.engine import (
    GaussNewtonIterations,
    LBFGScurvs,
    SimpleLBFGS,
)


class Extract:
    """Parse and extract minimiser setup from PHIL"""

    def __init__(
        self,
        master_phil,
        target,
        prediction_parameterisation,
        local_overrides="",
        cmdline_args=None,
        verbose=True,
    ):

        self._target = target
        self._prediction_parameterisation = prediction_parameterisation
        self._verbose = verbose

        arg_interpreter = command_line.argument_interpreter(master_phil=master_phil)

        user_phil = parse(local_overrides)
        cmdline_phils = []
        if cmdline_args:
            for arg in cmdline_args:
                cmdline_phils.append(arg_interpreter.process(arg))

        working_phil = master_phil.fetch(sources=[user_phil] + cmdline_phils)

        self._params = working_phil.extract().minimiser.parameters

        self.refiner = self.build_minimiser()

    def build_minimiser(self):

        assert self._params.engine in ["SimpleLBFGS", "LBFGScurvs", "GaussNewton"]

        if self._params.engine == "SimpleLBFGS":
            refiner = SimpleLBFGS(
                target=self._target,
                prediction_parameterisation=self._prediction_parameterisation,
                log=self._params.logfile,
            )
            return refiner

        if self._params.engine == "LBFGScurvs":
            refiner = LBFGScurvs(
                target=self._target,
                prediction_parameterisation=self._prediction_parameterisation,
                log=self._params.logfile,
            )
            return refiner

        if self._params.engine == "GaussNewton":

            refiner = GaussNewtonIterations(
                target=self._target,
                prediction_parameterisation=self._prediction_parameterisation,
                log=self._params.logfile,
            )
            return refiner



from __future__ import annotations

import itertools
import math
import pickle
import random

from libtbx.phil import parse

master_phil = parse(
    """
nrefl = 0
  .type = int
shoebox_size {
  x = 10
    .type = int
  y = 10
    .type = int
  z = 10
    .type = int
}
spot_size {
  x = 1.0
    .type = float
  y = 1.0
    .type = float
  z = 1.0
    .type = float
}
spot_offset {
  x = 0.0
    .type = float
  y = 0.0
    .type = float
  z = 0.0
    .type = float
}
rs_node_size = 0.05
  .type = float
rs_window_size = 0.25
  .type = float
mask_nsigma = 3.0
  .type = float
integrated_data_file = None
  .type = path
integrated_data_file_scale = 0.0
  .type = float
counts = 0
  .type = int
background = 0
  .type = int
background_a = 0.0
  .type = float
background_b = 0.0
  .type = float
background_c = 0.0
  .type = float
background_d = 0.0
  .type = float
pixel_mask = *all static precise
  .type = choice
background_method = *xds mosflm
  .type = choice
integration_methpd = *xds mosflm
  .type = choice
output {
  over = None
    .type = path
  under = None
    .type = path
  all = None
    .type = path
}
rotation {
  axis {
    x = 0.0
      .type = float
    y = 0.0
      .type = float
    z = 0.0
      .type = float
  }

  angle = 0.0
    .type = float
}
"""
)


def random_background_plane2(sbox, a, b, c, d):
    """Draw values from Poisson distribution for each position where the mean for
    that distribition is equal to a + b * i + c * j + d * k where a, b, c, d are
    floating point values and i, j, k are the shoebox indices in directions x, y
    and z respectively."""

    from scitbx.random import poisson_distribution, variate

    dz, dy, dx = sbox.focus()

    if b == c == d == 0.0:
        if a == 0:
            return
        g = variate(poisson_distribution(mean=a))
        for k in range(dz):
            for j in range(dy):
                for i in range(dx):
                    sbox[k, j, i] += next(g)
    else:
        for k in range(dz):
            for j in range(dy):
                for i in range(dx):
                    pixel = a + b * (i + 0.5) + c * (j + 0.5) + d * (k + 0.5)
                    if pixel != 0:
                        g = variate(poisson_distribution(mean=pixel))
                        sbox[k, j, i] += next(g)
    return


def random_background_plane(sbox, a, b, c, d):
    """Draw values from Poisson distribution for each position where the mean for
    that distribition is equal to a + b * i + c * j + d * k where a, b, c, d are
    floating point values and i, j, k are the shoebox indices in directions x, y
    and z respectively."""

    from scitbx.random import poisson_distribution, variate

    dz, dy, dx = sbox.focus()

    if b == c == d == 0.0:
        g = variate(poisson_distribution(mean=a))
        for k in range(dz):
            for j in range(dy):
                for i in range(dx):
                    sbox[k, j, i] += next(g)
    else:
        for k in range(dz):
            for j in range(dy):
                for i in range(dx):
                    pixel = a + b * i + c * j + d * k
                    g = variate(poisson_distribution(mean=pixel))
                    sbox[k, j, i] += next(g)
    return


def simple_gaussian_spots(params):
    from scitbx import matrix

    from dials.array_family import flex

    r = params.rotation
    axis = matrix.col((r.axis.x, r.axis.y, r.axis.z))
    if axis.length() > 0:
        rotation = axis.axis_and_angle_as_r3_rotation_matrix(r.angle, deg=True)
    else:
        rotation = matrix.sqr((1, 0, 0, 0, 1, 0, 0, 0, 1))

    # generate mask and peak values

    from dials.algorithms.shoebox import MaskCode

    mask_peak = MaskCode.Valid | MaskCode.Foreground
    mask_back = MaskCode.Valid | MaskCode.Background

    from dials.util.command_line import ProgressBar

    p = ProgressBar(title="Generating reflections")

    rlist = flex.reflection_table(params.nrefl)
    hkl = flex.miller_index(params.nrefl)
    s1 = flex.vec3_double(params.nrefl)
    xyzmm = flex.vec3_double(params.nrefl)
    xyzpx = flex.vec3_double(params.nrefl)
    panel = flex.size_t(params.nrefl)
    bbox = flex.int6(params.nrefl)

    for j in range(params.nrefl):
        p.update(j * 100.0 / params.nrefl)
        hkl[j] = (random.randint(0, 20), random.randint(0, 20), random.randint(0, 20))
        phi = 2 * math.pi * random.random()
        s1[j] = (0, 0, 0)
        xyzpx[j] = (0, 0, 0)
        xyzmm[j] = (0, 0, phi)
        panel[j] = 0
        bbox[j] = (
            0,
            params.shoebox_size.x,
            0,
            params.shoebox_size.y,
            0,
            params.shoebox_size.z,
        )

    p.finished("Generating %d reflections" % params.nrefl)
    intensity = flex.double(params.nrefl)
    shoebox = flex.shoebox(panel, bbox)
    shoebox.allocate_with_value(MaskCode.Valid)

    p = ProgressBar(title="Generating shoeboxes")

    for i in range(len(rlist)):

        p.update(i * 100.0 / params.nrefl)
        mask = shoebox[i].mask

        if params.pixel_mask == "precise":
            # flag everything as background: peak will me assigned later
            for j in range(len(mask)):
                mask[j] = mask_back
        elif params.pixel_mask == "all":
            # flag we have no idea what anything is
            mask_none = MaskCode.Valid | MaskCode.Foreground | MaskCode.Background
            for j in range(len(mask)):
                mask[j] = mask_none
        elif params.pixel_mask == "static":
            from scitbx.array_family import flex

            x0 = params.spot_offset.x + params.shoebox_size.x / 2
            y0 = params.spot_offset.x + params.shoebox_size.y / 2
            z0 = params.spot_offset.x + params.shoebox_size.z / 2
            sx = params.mask_nsigma * params.spot_size.x
            sy = params.mask_nsigma * params.spot_size.y
            sz = params.mask_nsigma * params.spot_size.z

            # The x, y, z indices
            z, y, x = zip(*itertools.product(*(range(n) for n in mask.all())))
            xyz = flex.vec3_double(flex.double(x), flex.double(y), flex.double(z))

            # Calculate SUM(((xj - xj0) / sxj)**2) for each element
            xyz0 = (x0, y0, z0)
            isxyz = (1.0 / sx, 1.0 / sy, 1.0 / sz)
            dxyz = sum(
                (x * isx) ** 2
                for x, isx in zip(((xyz - xyz0) * rotation).parts(), isxyz)
            )

            # Set the mask values
            index = dxyz <= 1.0
            index.reshape(mask.accessor())
            mask.set_selected(index, MaskCode.Valid | MaskCode.Foreground)
            mask.set_selected(not index, MaskCode.Valid | MaskCode.Background)

        sbox = shoebox[i].data

        # reflection itself, including setting the peak region if we're doing that
        # FIXME use flex arrays to make the rotation bit more efficient as this is
        # now rather slow...

        counts_true = 0
        for j in range(params.counts):
            _x = random.gauss(0, params.spot_size.x)
            _y = random.gauss(0, params.spot_size.y)
            _z = random.gauss(0, params.spot_size.z)

            Rxyz = rotation * matrix.col((_x, _y, _z)).elems

            x = int(Rxyz[0] + params.spot_offset.x + params.shoebox_size.x / 2)
            y = int(Rxyz[1] + params.spot_offset.y + params.shoebox_size.y / 2)
            z = int(Rxyz[2] + params.spot_offset.z + params.shoebox_size.z / 2)

            if x < 0 or x >= params.shoebox_size.x:
                continue
            if y < 0 or y >= params.shoebox_size.y:
                continue
            if z < 0 or z >= params.shoebox_size.z:
                continue
            sbox[z, y, x] += 1
            counts_true += 1
            if params.pixel_mask == "precise":
                mask[z, y, x] = mask_peak

        intensity[i] = counts_true

        if params.background:
            # background:flat;
            for j in range(params.background * len(sbox)):
                x = random.randint(0, params.shoebox_size.x - 1)
                y = random.randint(0, params.shoebox_size.y - 1)
                z = random.randint(0, params.shoebox_size.z - 1)
                sbox[z, y, x] += 1
        else:
            # or inclined
            random_background_plane(
                sbox,
                params.background_a,
                params.background_b,
                params.background_c,
                params.background_d,
            )

    rlist["miller_index"] = hkl
    rlist["s1"] = s1
    rlist["xyzcal.px"] = xyzpx
    rlist["xyzcal.mm"] = xyzmm
    rlist["bbox"] = bbox
    rlist["panel"] = panel
    rlist["shoebox"] = shoebox
    rlist["intensity.sum.value"] = intensity
    p.finished("Generating %d shoeboxes" % params.nrefl)

    return rlist


# FIXME to do:
#
#  - generate the reflections in reciprocal space w.r.t. a reciprocal space
#    node; will require taking a position then calculating a prediction for it
#    (i.e. when it will be in reflecting position) then extrapolate this and
#    add it to the appropriate voxel
#  - implement this as a flex array of events then perform this mapping in C++
#    to the target space, then accumulate into the shoebox in the target space
#    (can this be done in C++ as well?)
#  - apply the detector geometric correction due to the depth of the pixel, will
#    depend on the actual detector geometry and so on
#  - include input of beam, crystal, detector, scan, goniometer models so that
#    the predictions are realistic
#  - worth noting that if I can make this work it could form the kernel of old
#    deadly - taking all of the dials models as input along with some profile
#    parameters and generating an image of a spot...


def background_xds(rlist):
    # FIXME
    # from dials.extensions import XdsBackgroundExt
    # background = XdsBackgroundExt(None, None)
    # background.compute_background(rlist)

    # from math import erf, sqrt
    # from dials.algorithms.shoebox import MaskCode
    # from dials.array_family import flex
    # nsigma = 3
    # shoebox = rlist['shoebox']
    # for i in range(len(shoebox)):
    # data = shoebox[i].data
    # mask = shoebox[i].mask
    # bkgrd = shoebox[i].background

    # fg = flex.bool([m & MaskCode.Foreground != 0 for m in mask])
    # bg = flex.bool([m & MaskCode.BackgroundUsed != 0 for m in mask])
    # n = fg.count(True)
    # m = bg.count(True)
    # Sp = flex.sum(data.select(fg))
    # Bp = flex.sum(data.select(bg)) / m

    # F = 1.0 - erf(nsigma / sqrt(2.0))
    # c = 1.0 / (1 + F * n / m)
    # B = c * (Bp - (F / m) * Sp)
    ##   print F, c, Bp, n, m, Sp, B
    # print Bp, Sp
    # for j in range(len(bkgrd)):
    # bkgrd[j] = B

    return


def background_inclined(rlist):
    from dials.algorithms.background import InclinedSubtractor

    background = InclinedSubtractor()
    background(None, rlist)


def integrate_3d_summation(rlist):
    from dials.algorithms.integration.sum import sum_integrate_and_update_table

    sum_integrate_and_update_table(rlist)


def main(params):
    # first generate the reflections - this could be called from elsewhere
    rlist = simple_gaussian_spots(params)
    correct_intensities = list(rlist["intensity.sum.value"])
    del rlist["intensity.sum.value"]

    # now integrate those reflections using code from elsewhere
    if params.background_method == "xds":
        background_xds(rlist)
    elif params.background_method == "mosflm":
        assert params.pixel_mask != "all"
        background_inclined(rlist)

    integrate_3d_summation(rlist)

    integrated_intensities = list(rlist["intensity.sum.value"])

    # now scan through the reflection list and find those where the integration
    # gave an apparently duff answer i.e. outside of 3 sigma from correct value

    from dials.array_family import flex

    overestimates = []
    underestimates = []

    for j, (c, i) in enumerate(zip(correct_intensities, integrated_intensities)):
        sigma = math.sqrt(c)
        if math.fabs(c - i) < 3 * sigma:
            continue
        if i > params.counts:
            overestimates.append(j)
        else:
            underestimates.append(j)

    print(f"{len(overestimates)} overestimates, {len(underestimates)} underestimates")

    overestimates = rlist.select(flex.size_t(overestimates))
    underestimates = rlist.select(flex.size_t(underestimates))
    # now pickle these, perhaps

    if params.output.under:
        with open(params.output.under, "wb") as fh:
            pickle.dump(underestimates, fh, pickle.HIGHEST_PROTOCOL)
    if params.output.over:
        with open(params.output.over, "wb") as fh:
            pickle.dump(overestimates, fh, pickle.HIGHEST_PROTOCOL)
    if params.output.all:
        with open(params.output.all, "wb") as fh:
            pickle.dump(rlist, fh, pickle.HIGHEST_PROTOCOL)


if __name__ == "__main__":
    import sys

    from libtbx.phil import command_line

    cmd = command_line.argument_interpreter(master_params=master_phil)
    working_phil = cmd.process_and_fetch(args=sys.argv[1:])
    main(working_phil.extract())



from __future__ import annotations

import pkg_resources


class _Extension:
    """A base class for extension groups.
    This contains a common lookup mechanism and phil scope generator.
    """

    entry_point = None

    @classmethod
    def extensions(cls):
        """Return a list of all registered extension classes."""
        return [
            entry_point.load()
            for entry_point in pkg_resources.iter_entry_points(cls.entry_point)
        ]

    @classmethod
    def load(cls, name):
        """Get the requested extension class by name.

        :param name: The name of the extension
        :returns: The extension class
        """
        for entry_point in pkg_resources.iter_entry_points(cls.entry_point, name):
            # if there are multiple entry points with the same name then just return the first
            return entry_point.load()

    @classmethod
    def phil_scope(cls):
        """Get the phil scope for the interface or extension.

        :returns: The phil scope for the interface or extension
        """
        from libtbx.phil import parse

        if cls == _Extension:
            raise RuntimeError("Extension has no phil parameters")
        doc = "\n".join('"%s"' % d for d in cls.__doc__.strip().splitlines())
        master_scope = parse(f"{cls.name} .help={doc} {{}}")
        main_scope = master_scope.get_without_substitution(cls.name)
        assert len(main_scope) == 1
        main_scope = main_scope[0]
        if "phil" in cls.__dict__:
            main_scope.adopt_scope(cls.phil())

        def ext_names(extensions):
            names = []
            default_index = -1
            for ext in extensions:
                if "default" in ext.__dict__:
                    default_index = len(names)
                names.append(ext.name)
            if default_index < 0:
                default_index = 0
            if names:
                names[default_index] = "*" + str(names[default_index])
            return names

        exts = cls.extensions()
        if exts:
            algorithm = parse(
                f"""
        algorithm = {' '.join(ext_names(exts))}
          .help = "The choice of algorithm"
          .type = choice
      """
            )
            main_scope.adopt_scope(algorithm)
            for ext in exts:
                if hasattr(ext, "phil"):
                    help_str = "\n".join(
                        ['"%s"' % line for line in ext.__doc__.split()]
                    )
                    ext_master_scope = parse(f"{ext.name} .help={help_str} {{}}")
                    ext_phil_scope = ext_master_scope.get_without_substitution(ext.name)
                    assert len(ext_phil_scope) == 1
                    ext_phil_scope = ext_phil_scope[0]
                    ext_phil = ext.phil()
                    if ext_phil is not None:
                        ext_phil_scope.adopt_scope(ext.phil())
                        main_scope.adopt_scope(ext_master_scope)
        return master_scope


class SpotFinderThreshold(_Extension):
    """Extensions for threshold algorithms to be used in spot finding."""

    entry_point = "dials.spotfinder.threshold"
    name = "threshold"
    scope = "spotfinder"


class ProfileModel(_Extension):
    """
    The interface definition for a profile model.
    """

    entry_point = "dxtbx.profile_model"
    name = "profile"
    scope = "profile"


class Centroid(_Extension):
    """Extensions for centroid algorithms."""

    entry_point = "dials.integration.centroid"
    name = "centroid"
    scope = "integration"


class Background(_Extension):
    """Extensions for background algorithms."""

    entry_point = "dials.integration.background"
    name = "background"
    scope = "integration"



"""
Code to generate lookup values for algorithms/image/centroid/bias.h.
"""


from __future__ import annotations


def sum_of_erf(mu, sigma, N=1000):
    """
    Compute the sum of erf term

    :param mu: The Gaussian mean
    :param sigma: The Gaussian sigma
    :param N: The number of iterations in the sum
    """
    from math import erf, sqrt

    sum1 = 0
    sum2 = N * erf((N + 1 - mu) / (sqrt(2) * sigma))
    sum3 = N * erf((-N - mu) / (sqrt(2) * sigma))
    for i in range(-N, N):
        sum1 += erf((i + 1 - mu) / (sqrt(2) * sigma))
    return -0.5 * (sum1 + sum2 + sum3)


def compute_normal_bias_sq(sigma, N1=1000, N2=1000):
    """
    Compute the bias for a Gaussian profile

    :param sigma: The Gaussian sigma
    :param N1: The number of iterations for the sum_of_erf
    :param N2: The number of divisions for the integral
    """

    # If sigma is zero then we know the bias is 1/12.0
    if sigma == 0:
        return 1.0 / 12.0

    # The function to integrate
    def function(mu):
        return (sum_of_erf(mu, sigma, N1) - mu + 0.5) ** 2

    # Perform the numerical integration
    fa = function(0.0)
    fb = function(1.0)
    sum_f = 0.0
    for i in range(1, N2):
        sum_f += function(i * 1.0 / N2)
    return (1.0 / N2) * (fa / 2.0 + fb / 2.0 + sum_f)


def compute_lookup_table(max_sigma=0.5, N1=1000, N2=1000, N3=50):
    """
    Compute a lookup table of bias for Gaussian sigma values

    :param max_sigma: The maximum sigma to compute
    :param N1: The number of iterations for the sum_of_erf
    :param N2: The number of division for the integral
    :param N3: The number of elements in the table
    """
    sigma = []
    bias_sq = []
    for i in range(N3):
        s = i * max_sigma / N3
        b = compute_normal_bias_sq(s, N1, N2)
        sigma.append(s)
        bias_sq.append(b)

    return sigma, bias_sq


if __name__ == "__main__":

    sigma, bias_sq = compute_lookup_table()

    for s, b in zip(sigma, bias_sq):
        print(f"{s:0.2f} {b:0.7f}")



from __future__ import annotations

import dials_algorithms_centroid_simple_ext


def centroid(experiments, reflections, image_volume=None):
    """
    Do the centroiding

    :param experiments: The experiment list
    :param reflections: The reflection list
    """

    # Create a centroider instance
    centroider = dials_algorithms_centroid_simple_ext.Centroider()

    # Add all the experiments
    for exp in experiments:
        if exp.scan is not None:
            centroider.add(exp.detector, exp.scan)
        else:
            centroider.add(exp.detector)

    if image_volume is None:
        return centroider(reflections)
    return centroider(reflections, image_volume)
